"""
Carbon helper module - retrieve data from CryptoCompare
"""
__VERSION__ = "1.4"
__DATE__ = "01/Feb/2023"

import os as _os
import pandas as _pd
import hashlib as _hashlib
import requests as _requests
import pickle as _pickle
from collections import namedtuple as _namedtuple


pair_t = _namedtuple("pair", "tknb,tknq")

class CryptoCompare():
    """
    simple class formalizing interaction with the crypto compare API 
    
    :apikeyname:    the OS environment variable holding the API key
                    only used if no `apikey`; default is class.APIKEYNAME
    :apikey:        the API key; if True use without API key
    :datapath:      the path where all data is written (and read from)
    :raiseonerror:  if True, errors usually lead to an exception, otherwise to a None return
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    BASEURL = "https://min-api.cryptocompare.com" # must NOT end with /
    APIKEYNAME = "CCAPIKEY" # the name of the environment variable containing the API key
    RAISEONERROR = True
    DATAPATH = "cryptocompare"
    
    DEFAULT_TSYM = "usd"
    DEFAULT_LIMIT = 2000
    
    def __init__(self, apikeyname=None, apikey=None, raiseonerror=None):
        if raiseonerror is None:
            raiseonerror = self.RAISEONERROR
        self.raiseonerror = raiseonerror  
        if not (isinstance(apikey, str) or apikey is None or apikey is True):
            raise ValueError("apikey must be a string, None, or True", apikey)  
        if apikey is None:
            if apikeyname is None:
                apikeyname = self.APIKEYNAME
            apikey = _os.getenv(apikeyname)
            if apikey is None:
                print(f"Can't find API key {apikeyname} in environment variables.")
                print(f"Use `export {apikeyname}=<value>` to set it BEFORE you launch Jupyter")
                raise RuntimeError(f"API key not present. Use `export {apikeyname}=<value>` to set it before launching Jupyter.")
        self.apikey = apikey
    
    def url(self, endpoint):
        """
        returns the URL of a given endpoint
        """
        return f"{self.BASEURL}{endpoint}"
    
    @property
    def keydigest(self):
        """returns signature (=SHA1 hash) of the API key, or 0000... if anonymous"""
        if self.apikey is True:
            return "0"*40
        return _hashlib.sha1(self.apikey.encode()).hexdigest()
    
    def datafn(self, fn):
        """returns the full data file name, including path"""
        return _os.path.join(self.DATAPATH, fn)

    def cache(self, item):
        """
        reads a data item from the data cache
        """
        try:
            with open(self.datafn(f"{item}.pickle"), "rb") as f:
                result = _pickle.load(f)
        except:
            if not self.raiseonerror:
                return None
            raise
        return result
    
    def write_cache(self, item, data):
        """
        writes `data` to the cache under the name `item`
        
        :returns:    `item` on success, None (or raises) on failure
        """
        try:
            with open(self.datafn(f"{item}.pickle"), "wb") as f:
                _pickle.dump(data, f)
        except:
            if not self.raiseonerror:
                return None
            raise           
        return item
    
    QUERY_GET = "GET"
    QUERY_POST = "POST"
    def query(self, endpoint, params=None, method=None):
        """
        generic API query
        
        :endpoint:  the API endpoint to call, eg "/all/exchanges"
        :params:    the API parameters (parameters with value None will be removed)
        :method:    http method; default is QUERY_GET
        """
        if method is None:
            method = self.QUERY_GET
        if params is None:
            params = dict()
        url = self.url(endpoint)
        paramsq = {k:v for k,v in params.items() if not v is None}
        print("[query]", url, paramsq, f"[{str(self.keydigest)[:4]}]")
        if not self.apikey is True:
            paramsq["api_key"] = self.apikey
        
        if method == self.QUERY_GET:
            r = _requests.get(url, params=paramsq)
        elif method == self.QUERY_POST:
            raise ValueError("Method QUERY_POST has not been implemented yet.")
        else:
            raise ValueError("Unknown method. Use QUERY_XXX constants", method)
        
        if not r:
            if self.raiseonerror:
                raise RuntimeError(f"API query not successfull (status={r.status})", r)
            else:
                return None
        return r
    
    def query_allexchanges(self):
        """
        endpoint = /data/v4/all/exchanges
        
        https://min-api.cryptocompare.com/documentation?key=Other&cat=allExchangesV4Endpoint
        """
        r = self.query(
            endpoint="/data/v4/all/exchanges"
        )
        if r is None: return r
        return r.json().get("Data")

    
    def _cache_xxx(self, item, updatemethod, readonfail=True, updateonfail=False):
        """
        generic cached access
        
        :item:           the name of the item in the cache
        :updatemethod:   the method to call for updating it
        :readonfail:     if True, on cache miss updatemethod is called
        :updateonfail:   it True, on cache miss, updatemethod is called an item is 
                         written to cache
        """
        if updateonfail:
            readonfail = True
        try:
            return self.cache(item)
        except:
            print(f"[_cache_xxx] cache miss for item {item}")
            if readonfail:
                print(f"[_cache_xxx] reading {item} from API")
                data = updatemethod()
                if updateonfail:
                    print(f"[_cache_xxx] updating cache for {item} from API")
                    self.write_cache(item, data)
                return data
            else:
                if self.raiseonerror:
                    raise
                else:
                    return None
        
    def cache_allexchanges(self, readonfail=True, updateonfail=False):
        """cached access to query_allexchanges"""
        return self._cache_xxx(
            item="query_allexchanges", 
            updatemethod=self.query_allexchanges
        )
    
    def query_ratelimit(self):
        """
        endpoint = /stats/rate/limit
        
        https://min-api.cryptocompare.com/documentation?key=Other&cat=rateLimitEndpoint
        """
        r = self.query(
            endpoint="/stats/rate/limit"
        )
        if r is None: return r
        return r.json().get("Data")
        
    def query_coinlist(self):
        """
        endpoint = /data/all/coinlist
        
        https://min-api.cryptocompare.com/documentation?key=Other&cat=allCoinsWithContentEndpoint
        """
        r = self.query(
            endpoint="/data/all/coinlist"
        )
        if r is None: return r
        return r.json().get("Data")
    
    def cache_coinlist(self, readonfail=True, updateonfail=False):
        """cached access to query_coinlist"""
        return self._cache_xxx(
            item="query_coinlist", 
            updatemethod=self.query_coinlist
        )
        
    def query_indexlist(self):
        """
        endpoint = /data/index/list
        
        https://min-api.cryptocompare.com/documentation?key=Index&cat=listOfIndices
        """
        r = self.query(
            endpoint="/data/index/list"
        )
        if r is None: return r
        return r.json().get("Data")
    
    def cache_indexlist(self, readonfail=True, updateonfail=False):
        """cached access to query_indexlist"""
        return self._cache_xxx(
            item="query_indexlist", 
            updatemethod=self.query_indexlist
        )
    
    @staticmethod
    def ts_tocc(ts):
        """
        convert timestamp into format needed by CryptoCompare
        
        :ts:     the timestamp in any format that works for pd.Timestamp(ts)
        """
        return int(_pd.Timestamp(ts).timestamp())
    
    @staticmethod
    def ts_fromcc(ts):
        """
        convert timestamp from CryptoCompare format into pd.Timestamp format
        """
        return _pd.to_datetime(ts, unit='s', origin='unix') 
    
    FREQ_DAILY = "day"
    FD = FREQ_DAILY
    FREQ_HOURLY = "hour"
    FH = FREQ_HOURLY
    FREQ_MINUTELY = "minute"
    FM = FREQ_MINUTELY
    FREQS = (FREQ_DAILY, FREQ_HOURLY, FREQ_MINUTELY)
    def query_freqlypair(self, freq, fsym=None, tsym=None, e=None, limit=False, toTs=None, aspandas=True):
        """
        endpoints = /data/v2/histoday, /data/v2/histohour, /data/v2/histominute
        
        :freq:     FREQ_DAILY/FD, FREQ_HOURLY/FH, or FREQ_MINUTELY/FM
        :fsym:     cryptocurrency symbol of interest
        :tsym:     currency symbol to convert into
        :e:        exchange to obtain data from
        :limit:    number of data points to return (max: 2000; False defaults to that number)
        :toTs:     returns historical data BEFORE that timestamp
                   timestamp format either 1452680400 or pd.Timestamp compatible string 
        
        https://min-api.cryptocompare.com/documentation?key=Historical&cat=dataHistoday
        https://min-api.cryptocompare.com/documentation?key=Historical&cat=dataHistohour
        https://min-api.cryptocompare.com/documentation?key=Historical&cat=dataHistominute
        """
        if not freq in self.FREQS:
            raise ValueError("Unknow frequency {}. Use the FREQ_XXX constants provided.")
        endpoint = f"/data/v2/histo{freq}"
        params = {
                "fsym":     fsym,
                "tsym":     tsym if not tsym is None else self.DEFAULT_TSYM,
                "e":        e,
                "limit":    limit if not limit is False else self.DEFAULT_LIMIT,
                "toTs":     toTs,
        }
        r = self.query(endpoint=endpoint, params=params)
        if r is None: return r
        r_json = r.json()
        if r_json.get("Response") == "Error":
            if self.raiseonerror:
                raise RuntimeError("Query not successfull", r, r_json, endpoint, params)
            else:
                return None
        if not aspandas:
            return r_json().get("Data")
        try:   
            # print("[query_freqlypair]", endpoint, params, r)
            # print("[query_freqlypair] r", r_json())
            
            df = _pd.DataFrame.from_records(r_json["Data"]["Data"])
            df["datetime"] = [self.ts_fromcc(ts) for ts in df["time"]]
            df = df.set_index("datetime")
            del df["conversionType"]
            del df["conversionSymbol"]
            del df["time"]
            df = df[['open', 'close', 'high', 'low', 'volumefrom', 'volumeto']]
            return df
        except RuntimeError as e:
            if self.raiseonerror: 
                raise RuntimeError("Error {e}", endpoint, params, r)
            return None
    
    def query_dailypair(self, *args, **kwargs):
        """alias for query_freqlypair(FREQ_DAILY, ...)"""
        return self.query_freqlypair(self.FREQ_DAILY, *args, **kwargs)
    
    def query_hourlypair(self, *args, **kwargs):
        """alias for query_freqlypair(FREQ_HOURLY, ...)"""
        return self.query_freqlypair(self.FREQ_HOURLY, *args, **kwargs)

    def query_minutelypair(self, *args, **kwargs):
        """alias for query_freqlypair(FREQ_MINUTELY, ...)"""
        return self.query_freqlypair(self.FREQ_MINUTELY, *args, **kwargs)

    def ccycodes(self, symonly=True, fn=None):
        """
        returns information on currency codes
        
        :symonly:   if True (default) only return list of ccy symbold
        :fn:        the filename of the currency code file
        """
        if symonly:
            return self.join( self.unjoin(self.CCYCODES) )
        if fn is None:
            fn = _os.path.join(self.DATAPATH, "isoccy.csv")
        df = _pd.read_csv(fn, index_col=False)
        if symonly:
            symbols = list(set(df["Symbol"]))
            symbols.sort()
            return tuple(symbols)
        return df
    
    CCYCODES = """
    AED,AFN,ALL,AMD,ANG,AOA,ARS,AUD,AWG,AZN,BAM,BBD,BDT,BGN,BHD,BIF,BMD,
    BND,BOB,BOV,BRL,BSD,BTN,BWP,BYN,BZD,CAD,CDF,CHE,CHF,CHW,CLF,CLP,CNY,
    COP,COU,CRC,CUC,CUP,CVE,CZK,DJF,DKK,DOP,DZD,EGP,ERN,ETB,EUR,FJD,FKP,
    GBP,GEL,GHS,GIP,GMD,GNF,GTQ,GYD,HKD,HNL,HRK,HTG,HUF,IDR,ILS,INR,IQD,
    IRR,ISK,JMD,JOD,JPY,KES,KGS,KHR,KMF,KPW,KRW,KWD,KYD,KZT,LAK,LBP,LKR,
    LRD,LSL,LYD,MAD,MDL,MGA,MKD,MMK,MNT,MOP,MRU,MUR,MVR,MWK,MXN,MXV,MYR,
    MZN,NAD,NGN,NIO,NOK,NPR,NZD,OMR,PAB,PEN,PGK,PHP,PKR,PLN,PYG,QAR,RON,
    RSD,RUB,RWF,SAR,SBD,SCR,SDG,SEK,SGD,SHP,SLL,SOS,SRD,SSP,STN,SVC,SYP,
    SZL,THB,TJS,TMT,TND,TOP,TRY,TTD,TWD,TZS,UAH,UGX,USD,USN,UYI,UYU,UYW,
    UZS,VES,VND,VUV,WST,XAF,XAG,XAU,XCD,XDR,XOF,XPD,XPF,XPT,XSU,XUA,YER,
    ZAR,ZMW,ZWL
    """.strip()

    @staticmethod
    def join(tpl, sep=None):
        """join the tpl into comma separated strings"""
        if sep is None: sep = ", "
        return sep.join(str(s) for s in tpl)
        
    @staticmethod
    def unjoin(jstr, filter=None, sep=None):
        """
        unjoin the join string, stripping the result
        
        :jstr:      a (typically comma) separated string
        :filter:    filter to be applied (default: str)
        :sep:       the separator (default: comma)
        :returns:   tuple
        """
        if sep is None: sep = ","
        result = jstr.split(sep)
        if filter is None:
            filter = str
        result = ( filter(c.strip()) for c in result)
        return tuple(result)

    def aggr_query(self, 
            pairs, 
            fields=None, 
            incl_raw=True, 
            incl_raw_aggr=True, 
            incl_grand_aggr=True, 
            freq=None, **kwargs):
        """
        gets the data for pairs from the API and converts it into tables 
        
        :pairs:             the pairs to download, either comma separeted "ETH/USD, BTC/GBP, ..." 
                            or as tuple of tuples (("ETH", "USD"), ...)
        :fields:            the fields for which to create aggredate data frames, either comma separated
                            or as tuple/list; use FREQ_CLOSE and other FIELD_XXX constants here
        :incl_raw:          whether to include the individual raw data frames
        :incl_raw_aggr:     whether to include the aggregate raw data frame
        :incl_grand_aggr:   whether to include a grand aggregate (with double col name)
        :freq:              the data frequency [FREQ_DAILY (default), FREQ_HOURLY, FREQ_MINUTELY] 
        :kwargs:            passed through to `query_freqlypair` (eg `e`, `limit`, `toTs`)
        :returns:           dict with the results
        
        dict structure
        
        -gaggr
            - [data]
        -aggr
            -open
                - [data]
            -close
                - [data]
            ...
        -rawaggr
            - [data]
        -raw
            - "ETH/USD"
                - [data]
            ...
        """
        if fields is None:
            fields = self.FIELD_DEFAULT
        if isinstance(fields, str):
            fields = self.unjoin(fields)
        print("[aggr_query] fields", fields)
            
        if isinstance(pairs, str):
            pairs = tuple( self.pt_from_pair(p) for p in self.unjoin(pairs) )
        print("[aggr_query] pairs", pairs)
        
        if freq is None:
            freq = self.FREQ_DAILY
            
        result = {
            "gaggr": None,
            "aggr": None,
            "rawaggr": None,
            "raw": None,
        }
            
        print("[aggr_query] Querying for raw table", len(pairs))
        raw_tables = {
            (fsym, tsym): self.query_freqlypair(freq, fsym=fsym, tsym=tsym)
            for fsym, tsym in pairs
        }
        df_raw = _pd.concat(raw_tables, axis=1)
        result_raw = {self.pair_from_pt(p):v for p, v in raw_tables.items()}
        if incl_raw:
            result["raw"] = result_raw
        if incl_raw_aggr:
            result["rawaggr"] = _pd.concat(result_raw, axis=1)
        
        print("[aggr_query] Creating aggregate table")
        result["aggr"] = {
            field: self.reformat_raw_df(df_raw, field=field, dblcolnm=incl_grand_aggr) 
            for field in fields
        }
        if incl_grand_aggr:
            result["gaggr"] = _pd.concat(result["aggr"].values(), axis=1)
        return result
    
    @staticmethod
    def pairs_fields_from_df(df):
        """
        pairs and fields present in the dataframe
        
        :df:        data frame with index = (base token, quote token, field)
        :returns:   dict pairs: tuple( (tknp1, tnkq1), ...), fields: (field1, ...)
        """
        pairs  = ((tknb, tknq) for tknb, tknq, field in df.columns)
        pairs  = tuple(set(pairs))
        fields = (field for tknb, tknq, field in df.columns)
        fields = tuple(set(fields))
        return {"pairs": pairs, "fields": fields}

    FIELD_CLOSE = "close"
    FIELD_OPEN = "open"
    FIELD_HIGH = "high"
    FIELD_LOW = "low"
    FIELD_DEFAULT = FIELD_CLOSE
    @classmethod
    def reformat_raw_df(cls, df, field=None, dblcolnm=False):
        """
        reformats a raw df
        
        :df:        the raw df, as returned by a concatenation eg of daily_pair calls
        :field:     the name of the price field to use for the price
                    use FIELD_OPEN, FIELD_CLOSE etc; default: FIELD_DEFAULT
        :dblcolnm:  if True, the colname is (field, pair) instead of pair
        :returns:   the reformatted data frame
        """
        if field is None:
            field = cls.FIELD_DEFAULT

        if dblcolnm:
            result = (
                df[(*pair, field)].rename((field, f"{pair[0]}/{pair[1]}"), inplace=True)
                for pair in cls.pairs_fields_from_df(df)["pairs"]
            )
        else:
            result = (
                df[(*pair, field)].rename(f"{pair[0]}/{pair[1]}", inplace=True)
                for pair in cls.pairs_fields_from_df(df)["pairs"]
            )

        return _pd.concat(list(result), axis=1)

    @staticmethod
    def pt_from_pair(pair):
        """
        creates a pair tuple (tknb, tknq) from a pair 'TKNB/TKNQ'
        """
        return pair_t(*pair.split("/"))
    
    @staticmethod
    def pair_from_pt(pair_t):
        """
        creates a pair 'TKNB/TKNQ' from a pair tuple (tknb, tknq)
        """
        return "/".join(pair_t)

    @classmethod
    def coinlist(cls, coins, sep=",", aspt=False):
        """
        creates a coin list from separated string (does not touch lists)

        :coins:    either a string or a list/tuple
        :sep:      the separator of the string
        :aspt:     if True, result returned as pair tuple (using `pt_from_pair`)
        :returns:  original if not str; otherwise tuple of string or pr
        """
        f = cls.pt_from_pair if aspt else lambda x: x
        if isinstance(coins, str):
            return tuple(f(c.strip()) for c in coins.split(sep))
        else:
            return coins

    @classmethod
    def create_pairs(cls, coins, quotecoins=None):
        """
        create pair tuples from all possible combinations of coins and quotecoins
        
        :coins:        a list of coins, either ("tkn1", "tkn2") or "tkn1, tkn2"
        :quotecoins:   a list of quote coins; if None set equal to coins
        :returns:      all combinations as tuples (c, qc) with c!=qc
        """
        coins = cls.coinlist(coins)
        if quotecoins is None:
            quotecoins = coins
        else:
            quotecoins = cls.coinlist(quotecoins)
        result = ( (c,q) for q in quotecoins for c in coins)
        result = ( pair_t(c,q) for c,q in result if c != q)
        return tuple (result)

      


"""
wrapper classes for the node-based Carbon SDK via http
"""
__VERSION__ = "0.9"
__DATE__ = "11/Mar/2023"


from .sdkbase import SDKBase
from .carbonsdk0 import CarbonSDK0
from .carbonsdk import CarbonSDK
from .sdktoken import SDKToken, NULTKN, Tokens, TokenContainer




# ------------------------------------------------------------
# Auto generated test file `test_033_carbonpair.py`
# ------------------------------------------------------------
# source file   = NBTest_033_carbonpair.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 033
# test comment  = carbonpair
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, CarbonPair, P, __version__, __date__
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonPair))


#


#

# ------------------------------------------------------------
# Test      033
# File      test_033_carbonpair.py
# Segment   CarbonPair
# ------------------------------------------------------------
def test_carbonpair():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair=CarbonPair("ETH/USDC"))
    Sim
    
    assert (Sim.default_basetoken, Sim.default_quotetoken) == ('ETH', 'USDC')
    
    assert (Sim.tknb, Sim.tknq) == ('ETH', 'USDC')
    
    Sim.add_order("ETH", 10, 2000, 3000)["orders"]
    Sim.add_order("ETH", 10, 2000, 2500)["orders"]
    r = Sim.state()["orders"]
    assert list(r["pair"]) == ["ETHUSDC"]*4
    r
    
    r = Sim.amm_sells("ETH", 1)["trades"]
    assert list(r["p_unit"]) == ["USDC per ETH"]*3
    r
    
    Sim.state()["orders"]
    

# ------------------------------------------------------------
# Test      033
# File      test_033_carbonpair.py
# Segment   Decimals
# ------------------------------------------------------------
def test_decimals():
# ------------------------------------------------------------
    
    assert P("ETH/USDC").has_decimals == False
    assert P("ETH/USDC").sd().has_decimals == False
    assert P("ETH/USDC").sd(18).has_decimals == False
    assert P("ETH/USDC").sd(dec_tknq=18).has_decimals == False
    assert P("ETH/USDC").sd(18,6).has_decimals == True
    
    assert P("ETH/USDC").decimals == {'ETH': None, 'USDC': None, '_TKNB': None, '_TKNQ': None, '_DIFFQB': None}
    assert P("ETH/USDC").sd().decimals == {'ETH': None, 'USDC': None, '_TKNB': None, '_TKNQ': None, '_DIFFQB': None}
    assert P("ETH/USDC").sd(18,6).decimals == {'ETH': 18, 'USDC': 6, '_TKNB': 18, '_TKNQ': 6, '_DIFFQB': -12}
    
    DECDICT = {"ETH": 18, "USDC": 6, "WBTC": 8, "BNT": 18}
    
    assert P("ETH/USDC").sd(dct=DECDICT).decimals == {'ETH': 18, 'USDC': 6, '_TKNB': 18, '_TKNQ': 6, '_DIFFQB': -12}
    
    assert P("ETH/USDC").sd(dct=P.DECDICT).decimals["_DIFFQB"] == -12
    assert P("USDC/ETH").sd(dct=P.DECDICT).decimals["_DIFFQB"] == 12
    assert P("SHIB/ETH").sd(dct=P.DECDICT).decimals["_DIFFQB"] == 0
    assert P("WBTC/ETH").sd(dct=P.DECDICT).decimals["_DIFFQB"] == 10
    assert P("NEAR/ETH").sd(dct=P.DECDICT).decimals["_DIFFQB"] == 18
    
    assert P("ETH/USDC").sd(dct=P.DECDICT).decdiffqb == -12
    assert P("USDC/ETH").sd(dct=P.DECDICT).decdiffqb == 12
    assert P("SHIB/ETH").sd(dct=P.DECDICT).decdiffqb == 0
    assert P("WBTC/ETH").sd(dct=P.DECDICT).decdiffqb == 10
    assert P("NEAR/ETH").sd(dct=P.DECDICT).decdiffqb == 18
    
    ETHUSDC = P("ETH/USDC").sd(18,6)
    assert str(ETHUSDC) == "P('ETH/USDC').sd(18,6)"
    assert P(ETHUSDC).has_decimals
    assert str(P(ETHUSDC)) == "P('ETH/USDC').sd(18,6)"
    
    


from .sdktoken import Tokens
from .sdkbase import SDKBase

n = SDKBase.c2s

class CarbonSDK0(SDKBase):
    """
    mostly 1:1 wrapper around the http interface to the carbon SDK
    """
    URL = "http://localhost:3118" # C=3, A=1, R=18 (Carbon)
    TOKEN = "carbontoken"
    APIBASE = "api"
    SYNCDEFAULT = False

    ########################################
    ## Test API requests
    def mul(self, a, b, sync=None):
        "multiplies a,b, plus random wait"
        return self.call("mul", params={"a": a, "b": b}, sync=sync)

    def plus(self, a, b, sync=None):
        "adds a,b, plus random wait"
        return self.call("plus", params={"a": a, "b": b}, sync=sync)
    
    def qmul(self, a, b, sync=None):
        "multiplies a,b, no wait"
        return self.call("qmul", params={"a": a, "b": b}, sync=sync)

    def qplus(self, a, b, sync=None):
        "adds a,b, no wait"
        return self.call("qplus", params={"a": a, "b": b}, sync=sync)

    ########################################
    ## ERC20 API requests
    def getERC20Balance(self, tokens, address):
        """
        calls the ERC20 balanceOf method for each token in tokens

        :tokens:    list of SDKToken objects
        :address:   address to check balance of
        :returns:   dict of balances in token units
        """
        r = self.req("erc20/balance_of", params={
                "tokens": [t.a for t in tokens],
                "address": address,
            })
        data = self._checkresult(r)
        bal = (self.bn2int(bal) for bal in data)
        result = {t.T: t.w2t(b) for t,b in zip(tokens, bal)}
        return result
    
    def getERC20Decimals(self, tokens):
        """
        calls the ERC20 decimals method for each token in tokens

        :tokens:    list of SDKToken objects
        :returns:   dict
        """
        r = self.req("erc20/decimals", params={
                "tokens": [t.a for t in tokens],
            })
        data = self._checkresult(r)
        result = {t.T: d for t,d in zip(tokens, data)}
        return result

    def getERC20Symbol(self, tokens):
        """
        calls the ERC20 symbol method for each token in tokens

        :tokens:    list of SDKToken objects or addresses
        :returns:   dict
        """
        try:
            tokensa = [t.a for t in tokens]
            tokenst = [t.T for t in tokens]
        except:
            tokensa = tokens
            tokenst = tokens

        r = self.req("erc20/symbol", params={
                "tokens": tokensa,
            })
        data = self._checkresult(r)
        result = {t: d for t,d in zip(tokenst, data)}
        return result
    
    ########################################
    ## Call API requests
    def getMatchActions(self, amountWei, tradeByTargetAmount, orders, sync=None):
        """
        TODO: docstring

        :amountWei:             amount of the trade in Wei  
        :tradeByTargetAmount:   whether the trade is by target amount (True) or by source amount (False) 
        :orders:                list of orders to match against
        :sync:                  whether to call sync or async
        """
        return self.call(n("getMatchActions"), params={
                "amountWei": str(amountWei), 
                "tradeByTargetAmount": bool(tradeByTargetAmount), 
                "orders": orders
            }, sync=sync)
    
    # def startDataSync(self, sync=None):
    #     """
    #     fetches blockchain data and subscribes to events

    #     :sync:      whether to call sync or async (should only be called async)
    #     :returns:   a promise that resolves when the data sync is complete
        
    #     This method should be called before any other actions are taken with the SDK,
    #     with the exception of strategy creation, update and delete that don't rely 
    #     on cached data. The promise returned resolves when the data sync is complete.
    #     """
    #     return self.call(n("startDataSync"), params={}, sync=sync)

    # def isInitialized(self, sync=None):
    #     """
    #     whether or not the SDK has been initialized, ie whether startDataSync has completed

    #     :sync:      whether to call sync or async (should be called sync)
    #     :returns:   whether the SDK has been initialized as a boolean
    #     """
    #     return self.call(n("isInitialized"), params={}, sync=sync)
    
    def addr(self):
        """
        returns the wallet address of the server

        :returns:   the wallet address of the server as a string
        """
        r = self.req(n("addr"))
        rj = r.json()
        if rj["success"]:
            return rj["data"]
        else:
            return None

    def pairs(self, inclstr=False):
        """
        all token pairs present in the Carbon contract

        :inclstr:   whether to include the string representation of the pairs
        :returns:   token pairs as an array of token addresses if self.Tokens is None, 
                    else as an list of pairs of SDKToken objects if inclstr is False,
                    else also a string of pairs of token symbols

        """
        T = self.Tokens
        r = self.call(n("pairs"), params={}, sync=True)
        pairs_raw = self._checkresult(r)
        if Tokens is None:
            return pairs_raw
        pairs = [ [Tokens(p[0]), Tokens(p[1])] for p in pairs_raw]
        if not inclstr:
            return pairs
        pairs_s = ", ".join([f"{p[0].T}/{p[1].T}" for p in pairs])
        return pairs, pairs_s
        
    def hasLiquidityByPair(self, sourceTokenAddr=None, targetTokenAddr=None, pairs=None):
        """
        whether a specfic pair has any open liquidity positions

        :sourceTokenAddr:       source token* (as address)
        :targetTokenAddr:       target token* (as address)
        :pairs:                 pairs of dicts {"sourceToken":..., "targetToken":...} of token addresses**
        :returns:               whether the directed pair has liquidity (boolean)

        *source and target are seen from the point of view of the trader, not the AMM/LP
        **either pairs or sourceTokenAddr and targetTokenAddr must be provided, but not both
        """
        if not pairs is None:
            if not (sourceTokenAddr is None and targetTokenAddr is None):
                raise ValueError("ERROR: either pairs or sourceTokenAddr and targetTokenAddr must be provided, but not both")
            r = self.call(n("hasLiquidityByPair"), params={
                "pairList": [{k: str(v) for k,v in p.items()} for p in pairs], 
            }, sync=True)
        else:
            r = self.call(n("hasLiquidityByPair"), params={
                    "sourceToken": str(sourceTokenAddr), 
                    "targetToken": str(targetTokenAddr)
                }, sync=True)
        return self._checkresult(r)
        
    
    def getLiquidityByPair(self, sourceTokenAddr, targetTokenAddr):
        """
        liquidity value for a specific pair

        :sourceTokenAddr:       source token* (as address)
        :targetTokenAddr:       target token* (as address)
        :sync:                  whether to call sync or async (should be called sync)
        :returns:               liquidity value in units of the target token as a float
        
        *source and target are seen from the point of view of the trader, not the AMM/LP
        """
        r = self.call(n("getLiquidityByPair"), params={
                "sourceToken": str(sourceTokenAddr), 
                "targetToken": str(targetTokenAddr)
            }, sync=True)
        data = self._checkresult(r)
        return float(data)
       
    def getMinRateByPair(self, sourceTokenAddr, targetTokenAddr):
        """
        the minimum exchange rate offered on Carbon between those two tokens

        :sourceTokenAddr:   source token* (as address; also base token)
        :targetTokenAddr:   target token* (as address; also quote token)
        :returns:           minimum rate (target token per source token)

        *source and target are seen from the point of view of the trader, not the AMM/LP
        """
        r = self.call(n("getMinRateByPair"), params={
                "sourceToken": str(sourceTokenAddr), 
                "targetToken": str(targetTokenAddr)
            }, sync=True)
        data = self._checkresult(r)
        return float(data)

    def getMaxRateByPair(self, sourceTokenAddr, targetTokenAddr):
        """
        gets the maximum exchange rate offered on Carbon between those two tokens

        :sourceTokenAddr:   source token* (as address; also base token)
        :targetTokenAddr:   target token* (as address; also quote token)
        :returns:           maximum rate (target token per source token)
        
        *source and target are seen from the point of view of the trader, not the AMM/LP
        """
        r = self.call(n("getMaxRateByPair"), params={
                "sourceToken": str(sourceTokenAddr), 
                "targetToken": str(targetTokenAddr)
            }, sync=True)
        data = self._checkresult(r)
        return float(data)
        
    def getRateLiquidityDepthByPair(self, sourceTokenAddr, targetTokenAddr, rate):
        """
        liquidity depth (ie cumulative liquidity) at a given exchange rate

        :sourceTokenAddr:   source token* (as address; also base token)
        :targetTokenAddr:   target token* (as address; also quote token)
        :rate:              rate (target token per source token)
        :returns:           liquidity depth, ie cumulative liquidity at rate or better for trader
                            measured in units of the target token**
        
        *source and target are seen from the point of view of the trader, not the AMM/LP
        **the liqudity function is a decreasing function of the rate; at max rate it is
        at the value returned by getLiquidityByPair, and at min rate it is 0
        """
        try:
            rate = [str(r) for r in rate]
        except:
            rate = str(rate)
        #print("[getRateLiquidityDepthByPair}", rate)
    
        r = self.call(n("getRateLiquidityDepthByPair"), params={
                "sourceToken": str(sourceTokenAddr), 
                "targetToken": str(targetTokenAddr), 
                "rate": rate
            }, sync=True)
        data = self._checkresult(r)
        try:
            data = float(data)
        except:
            data = [float(d) if not isinstance(d, dict) else None for d in data]
        return data
        
    def getUserStrategies(self, user):
        """
        gets the strategies that are owned by the user

        :user:      user (as address)
        :returns:   a list of strategy objects
        """
        r = self.call(n("getUserStrategies"), params={
                "user": str(user)
            }, sync=True)
        return self._checkresult(r)

    
    ########################################################################
    ## main trade functions (taker actions)
    MAXDEADLINE = str(2**32-1)
    MAXMAXINPUT = str(2**128-1) # TODO: wrong; maxInput is a token resultion number (string)
    
    def getTradeData(self, sourceToken, targetToken, amount, tradeByTargetAmount, filter=None):
        """
        Get the data needed to process a trade (trade action step 1)

        :sourceToken:               source token (as address)
        :targetToken:               target token (as address)
        :amount:                    amount of the trade
        :tradeByTargetAmount:       whether the trade, as seen by the trader, is by 
                                    target amount (True) or by source amount (False)
        :filter:                    filter to apply to the orders
        :returns:                   trade data dict
        """
        if filter is None: filter = []
        r = self.call(n("getTradeData"), params={
                "sourceToken": str(sourceToken), 
                "targetToken": str(targetToken), 
                "amount": str(amount), 
                "tradeByTargetAmount": bool(tradeByTargetAmount), 
                "filter": filter
            }, sync=True)
        return self._checkresult(r)

    def composeTradeByTargetTransaction(self, sourceToken, targetToken, tradeActions=None, deadline=None, maxInput=None, overrides=None):
        """
        Composes a trade by target transaction (trade action step 2)

        :sourceToken:               source token (as address)
        :targetToken:               target token (as address)
        :tradeActions:              list of actions to apply to the orders
        :deadline:                  deadline of the trade (block number as string; None=unlimited)
        :maxInput:                  maximum input amount (as string; None=unlimited)
        :overrides:                 overrides to apply to the transaction
        """
        if deadline is None: deadline = self.MAXDEADLINE
        if maxInput is None: maxInput = self.MAXMAXINPUT
        if overrides is None: overrides = dict()
        if not overrides.get("gasLimit"): overrides["gasLimit"] = 999999999
        if tradeActions is None: tradeActions = []
        tradeActions = [self.int2strd(d) for d in tradeActions]
        r = self.call(n("composeTradeByTargetTransaction"), params={
                "sourceToken": str(sourceToken), 
                "targetToken": str(targetToken), 
                "tradeActions": tradeActions, 
                "deadline": str(deadline), 
                "maxInput": str(maxInput), 
                "overrides": overrides
            }, sync=True)
        return self._checkresult(r)

    def composeTradeBySourceTransaction(self, sourceToken, targetToken, tradeActions, deadline=None, minReturn=None, overrides=None):
        """
        Composes a trade by source transaction (trade action step 2)

        :sourceToken:               source token (as address)
        :targetToken:               target token (as address)
        :tradeActions:              actions to apply to the orders
        :deadline:                  deadline of the trade (block number as string; None=unlimited)
        :minReturn:                 minimum return amount (as string; None=unlimited)
        :overrides:                 overrides to apply to the transaction
        :sync:                      whether to call sync or async
        """

        if deadline is None: deadline = self.MAXDEADLINE
        if minReturn is None: minReturn = 1
        if overrides is None: overrides = dict()
        if not overrides.get("gasLimit"): overrides["gasLimit"] = 999999999
        if tradeActions is None: tradeActions = []
        tradeActions = [self.int2strd(d) for d in tradeActions]
        r = self.call(n("composeTradeBySourceTransaction"), params={
                "sourceToken": str(sourceToken), 
                "targetToken": str(targetToken), 
                "tradeActions": tradeActions, 
                "deadline": str(deadline), 
                "minReturn": str(minReturn), 
                "overrides": overrides
            }, sync=True)
        return self._checkresult(r)


    #######################################################
    ## other trade functions (taker actions)
    def getMatchParams(self, sourceToken, targetToken, amount, tradeByTargetAmount, sync=None):
        """
        returns the data needed to process a trade

        :sourceToken:               source token (as address)
        :targetToken:               target token (as address)
        :amount:                    amount of the trade
        :tradeByTargetAmount:       whether the trade, as seen by the trader, is by 
                                    target amount (True) or by source amount (False)
        :sync:                      whether to call sync or async
        :returns:                   ??? TODO

        this function returns the data for a given source and target token pair; this data
        can then be used to call matchBySourceAmount or matchByTargetAmount; to finally get 
        the actions those will be passed to getTradeDataFromActions
        """
        return self.call(n("getMatchParams"), params={
                "sourceToken": str(sourceToken), 
                "targetToken": str(targetToken), 
                "amount": str(amount), 
                "tradeByTargetAmount": bool(tradeByTargetAmount)
            }, sync=sync)
    
    def getTradeDataFromActions(self, sourceToken, targetToken, tradeByTargetAmount, actionsWei=None, sync=None):
        """
        Creates an unsigned transaction to fulfill a trade using an array of trade actions

        :sourceToken:               source token (as address)
        :targetToken:               target token (as address)
        :tradeByTargetAmount:       whether the trade, as seen by the trader, is by 
                                    target amount (True) or by source amount (False)
        :actionsWei:                actions to apply to the orders in Wei
        :sync:                      whether to call sync or async
        :returns:                   ??? TODO
        """
        if actionsWei is None: actionsWei = []
        actionsWei = [self.int2strd(d) for d in actionsWei]
        return self.call(n("getTradeDataFromActions"), params={
                "sourceToken": str(sourceToken), 
                "targetToken": str(targetToken), 
                "tradeByTargetAmount": bool(tradeByTargetAmount), 
                "actionsWei": actionsWei
            }, sync=sync)

    ########################################################################
    ## maker actions
    def createBuySellStrategy(self, baseToken, quoteToken, buyPriceLow, buyPriceHigh, buyBudget, sellPriceLow, sellPriceHigh, sellBudget, overrides=None):
        """
        Creates a Carbon Strategy, composed of one buy order and one sell order

        :baseToken:         base token (as address; base token = risk asset)
        :quoteToken:        quote token (as address; quote token = numeraire asset)
        :buyPriceLow:       low buy price (in quote token per base token)
        :buyPriceHigh:      high buy price (ditto)
        :buyBudget:         buy budget (in quote token)
        :sellPriceLow:      low sell price (ditto)
        :sellPriceHigh:     high sell price (ditto)
        :sellBudget:        sell budget (in base token)
        :overrides:         overrides to apply to the transaction
        :sync:              whether to call sync or async
        :returns:           unsigned transaction that can be used to create the strategy
        """
        if overrides is None: overrides = dict()
        if not overrides.get("gasLimit"): overrides["gasLimit"] = 999999999
        r = self.call(n("createBuySellStrategy"), params={
                "baseToken": str(baseToken), 
                "quoteToken": str(quoteToken), 
                "buyPriceLow": str(buyPriceLow), 
                "buyPriceHigh": str(buyPriceHigh), 
                "buyBudget": str(buyBudget), 
                "sellPriceLow": str(sellPriceLow), 
                "sellPriceHigh": str(sellPriceHigh), 
                "sellBudget": str(sellBudget), 
                "overrides": overrides
            }, sync=True)
        return self._checkresult(r)

    def updateStrategy(self, strategyId, encoded, baseToken, quoteToken, update, buyMarginalPrice, sellMarginalPrice, overrides=None, sync=None):
        """
        updates an existing strategy

        :strategyId:            strategy id
        :encoded:               encoded strategy ??? TODO
        :baseToken:             base token (as address)
        :quoteToken:            quote token (as address)
        :update:                update to apply to the strategy
        :buyMarginalPrice:      buy marginal price (in quote token per base token)
        :sellMarginalPrice:     sell marginal price (ditto)
        :overrides:             overrides to apply to the transaction
        :sync:                  whether to call sync or async
        :returns:               ??? TODO
        """
        if overrides is None: overrides = dict()
        if not overrides.get("gasLimit"): overrides["gasLimit"] = 999999999
        return self.call(n("updateStrategy"), params={
                "strategyId": str(strategyId), 
                "encoded": encoded, 
                "baseToken": str(baseToken), 
                "quoteToken": str(quoteToken), 
                "update": bool(update), 
                "buyMarginalPrice": str(buyMarginalPrice), 
                "sellMarginalPrice": str(sellMarginalPrice), 
                "overrides": overrides
            }, sync=sync)
    
    def deleteStrategy(self, strategyId, overrides=None):
        """
        deletes a strategy

        :strategyId:        strategy id
        :overrides:         overrides to apply to the transaction
        :returns:           ??? TODO
        """
        if overrides is None: overrides = dict()
        if not overrides.get("gasLimit"): overrides["gasLimit"] = 999999999
        r = self.call(n("deleteStrategy"), params={
                "strategyId": str(strategyId), 
                "overrides": overrides,
            }, sync=True)
        return self._checkresult(r)   
    
    ########################################
    ## Non-call API requests
    def version(self):
        """calls the root endpoint of the API server (version)"""
        r = self.req0("/")
        return r.json()

    class UnknownReqIdError(ValueError): pass
    def result(self, reqid):
        """
        retrieves the result corresponding to reqid [sync]
        
        :returns:   return value depends on the state encapsulated in the API result
                    :available:     if the result of the API call is available then the result
                                    is returned; note that some of the returned API data is 
                                    removed (specifially, only r["data"]["result"] returned)
                    :awaiting:      if the result of the API call is still awaiting returns True
                    :unknown:       if reqid is unknown, returns False (or raises UnknownReqIdError)
        """
        try:
            url = self.join("result", str(reqid))
            r = self.req(url)
            rj = r.json()
        except:
            raise self.APICallError("Error calling result API", url, r)
        if rj["success"]:
            return rj["data"]["result"]
        if rj["awaiting"] is True:
            return True
        if self.raiseonerror:
            raise self.UnknownReqIdError("Unknown request id", reqid)
        return False
    
    def signsubmittx(self, tx, sign=True):
        """
        signs and submits a transaction (sync only)
        
        :tx:        the transaction to be submitted (signed or unsigned)
        :sign:      whether to sign the transaction (True=sign)
        :returns:   the transaction receipt
        """
        r = self.req(n("signsubmittx"), params={"tx": tx, sign: bool(sign)})
        return self._checkresult(r)
    
    ########################################
    ## Helper functions
    def _checkresult(self, r):
        """
        checks the result of a call to the API and returns the data or raises an error

        :r:     the request objects returns by the call to the API
        """
        if not r.status_code == 200:
            raise RuntimeError(f"ERROR (status = {r.status_code}): {r}")
        try:
            rj = r.json()
        except:
            raise RuntimeError(f"ERROR: can't convert to json {r.body}")
        try:
            return rj["data"]
        except:
            raise RuntimeError(f"ERROR: {rj}")
 

    def reformatStrategy(self, strat, nsd=True):
        """
        converts the values of a strategy (`createBuySellStrategy`) into Python-native objects
        
        :strat:   the strategy object returned from the API
        :nsd:     number significant digits (True=default; None=no rounding)
        """
        T = self._assertTokensProvided()
        if nsd is True: nsd = 8
        s2 = dict({"encoded": dict()})
        for f in ("buyPriceLow", "buyPriceHigh", "buyBudget", "sellPriceLow", "sellPriceHigh", "sellBudget"):
            s2[f] = float(strat[f]) if nsd is None else self.roundsd(float(strat[f]), nsd)
        s2["id"] = s2["encoded"]["id"] = self.bn2int(strat["id"])
        for f in ("baseToken", "quoteToken"):
            s2[f] = self.Tokens.byaddr1(strat[f], tknonly=True)
        for f in ("token0", "token1"):
            s2["encoded"][f] = self.Tokens.byaddr1(strat["encoded"][f], tknonly=True)
        for o in ("order0", "order1"):
            s2["encoded"][o] = dict()
            for f in ("y", "z", "A", "B"):
                s2["encoded"][o][f] = self.bn2int(strat["encoded"][o][f])
        return s2



"""
class representing an ERC20 token, in particular holding the token name and address
"""
__VERSION__ = "1.0"
__DATE__ = "8/Mar/2023"

from dataclasses import dataclass

@dataclass
class SDKToken():
    """
    describes a token for the Carbon SDK

    :token:     token ticker (eg "ETH")
    :address:   token address ("0x...")
    :decimals:  number of decimals (default None is 18)
    :name:      token name (default None is token)
    """    
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    token: str
    address: str
    decimals: int = None
    name: str = None
    
    def __post_init__(self):
        if not self.decimals is None:
            self.decimals = int(self.decimals)
        self.token = self.token.strip().upper()
        self.address = str(self.address).strip().lower()
        try:
            int(self.address, 16)
        except ValueError:
            raise ValueError(f"address {self.address} is not a hex number")
        if self.name is None: self.name = self.token

    @property
    def a(self):
        """alias for self.address"""
        return self.address
    
    @property
    def T(self):
        """alias for self.token"""
        return self.token 
    
    @property
    def d(self):
        """alias for self.decimals"""
        return self.decimals

    def w2t(self, wei):
        """converts wei to token"""
        return wei / 10**self.decimals
    
    def t2w(self, tokenmt):
        """converts tokenmt to wei"""
        return int(tokenmt * 10**self.decimals)
    
    @property
    def isnul(self):
        """returns True if token is NULTKN"""
        return self.token == "NULTKN"
    
    def __str__(self):
        return f"{self.token} ({self.address})"
    
    def __iter__(self):
        return iter((self.token, self.address, self.decimals, self.name))

NULTKN = SDKToken("NULTKN", "0x0000000000000000000000000000000000000000") 

class TokenContainer():
    """
    container class for SDKToken objects

    :tokens:        list of SDKToken objects
    :raiseonerror:  if True, raises KeyError if token not found; otherwise returns NULTKN
    """
    VERSION = __VERSION__
    DATE = __DATE__

    def __init__(self, tokens, raiseonerror=None):
        if raiseonerror is None: raiseonerror = True
        self.raiseonerror = raiseonerror
        self._byticker = {t.token: t for t in tokens}
        self._byaddr   = {t.address: t for t in tokens}
        self._tickers  = set(self._byticker.keys())
        if not len(self._tickers) == len(self._byaddr):
            raise ValueError(f"{len(self._tickers) - len(self._byaddr)} duplicate token tickers found", self._tickers, self._byaddr)

    @classmethod
    def fromlist(cls, tokens, raiseonerror=None):
        """creates a TokenContainer from a list of token parameters"""
        return cls([SDKToken(*tuple(t)) for t in tokens], raiseonerror)
    
    @classmethod
    def fromcsv(cls, csv, raiseonerror=None):
        """creates a TokenContainer from a csv string"""
        return cls.fromlist([t.split(",") for t in csv.strip().splitlines()], raiseonerror)
    
    @classmethod
    def restore(cls, ticker_or_addr_or_tkn, raiseonerror=None):
        """converts a ticker or address to an SDKToken from the TokenContainer (SDKToken not touched)"""
        if isinstance(ticker_or_addr_or_tkn, SDKToken):
            return ticker_or_addr_or_tkn
        try:
            tkn = cls.byaddr(ticker_or_addr_or_tkn, raiseonerror=True)
        except:
            try:
                tkn = cls.byticker(ticker_or_addr_or_tkn, raiseonerror=True)
            except:
                if raiseonerror:
                    raise ValueError(f"ticker or address {ticker_or_addr_or_tkn} not found in container")
                else:
                    tkn = None
        return tkn
    
    def byticker(self, token, raiseonerror=None, returnnultkn=True):
        """gets a token from the container (by ticker)"""
        if raiseonerror is None: raiseonerror = self.raiseonerror
        token = token.upper()
        try:
            return self._byticker[token]
        except KeyError:
            if raiseonerror:
                raise KeyError(f"token {token} not found in container")
            else:
                if returnnultkn:
                    return NULTKN
                else:
                    return None
    
    def byaddr(self, addr, raiseonerror=None, returnnultkn=True):
        """
        gets a token from the container (by address)
        
        :addr:          token address as hex string
        :raiseonerror:  if True, raises KeyError if token not found; otherwise returns nul value
        :returnnultkn:  if True, returns NULTKN if token not found; otherwise returns None
        """
        if raiseonerror is None: raiseonerror = self.raiseonerror
        addr = addr.lower()
        try:
            return self._byaddr[addr]
        except KeyError:
            if raiseonerror:
                raise KeyError(f"address {addr} not found in container") 
            else:
                if returnnultkn:
                    return NULTKN
                else:
                    return None
            
    def byaddr1(self, addr, tknonly=False):
        """
        gets a token from the container (by address); keeps addr if not found
        
        :addr:      token address as hex string
        :tknonly:   if True, returns only the token name; otherwise returns object
        :returns:   token name or object, depending on tknonly
        """
        addr = addr.lower()
        try:
            tkno = self._byaddr[addr]
            return tkno.T if tknonly else tkno
        except KeyError:
            return addr
    
    def containsall(self, tokenListorSet):
        """checks if all tokens in the list are in the container"""
        return tokenListorSet.issubset(self._tickers)

    @property
    def _all(self):
        """gets the full list of tokens from the container"""
        return tuple(self._byticker.values())
    
    def __getitem__(self, token):
        return self.byticker(token)
    
    def __getattr__(self, token):
        return self.byticker(token)
    
    def __len__(self):
        return len(self._byticker)
    
    def __iter__(self):
        return iter(self._all)
    
    def __repr__(self):
        return f"{self.__class__.__name__}({self._all})"
    
    def __hash__(self):
        return int(self.address, 16)
    
    def __str__(self):
        return str(tuple(self._byticker.keys()))
    
    def __call__(self, addr):
        """alias for self.byaddr1()"""
        return self.byaddr1(addr=addr)

# note: use the script below to generate the csv string

# first run this in Node.js to get the token addresses
# import { ethers } from 'ethers';
# const rpcUrl = process.env.ALCHEMY_RPCURL;
# const provider = new ethers.providers.JsonRpcProvider(rpcUrl);
# const abi = ['function decimals() view returns (uint8)']; // replace with the ABI of the ERC20 contract
# const tokens = ['0x7fc66500c84a76ad7e9c93437bfc5ac33e2ddae9',...]; // replace with the addresses of the ERC20 contracts
# var   tokenDecimals = {}
# for (var i = 0; i < tokens.length; i++) {
#     let taddr = tokens[i]
#     let tcontract = new ethers.Contract(taddr, abi, provider);
#     let tdecimals
#     try{
#         tdecimals = await tcontract.decimals()
#     } catch (e) {
#         tdecimals = "error"
#     }
#     console.log(taddr, tdecimals)
#     tokenDecimals[taddr] = tdecimals
# }
# console.log(tokenDecimals)
#
# then run this in Python to generate the csv string
# decs = <output from console.log(tokenDecimals)>
# for t in Tokens:
#    print(f"{t.token}, {t.a}, {decs[t.a]}")
# ETH of course need to be adjusted manually to 18 decs

Tokens = TokenContainer.fromcsv("""
AAVE, 0x7fc66500c84a76ad7e9c93437bfc5ac33e2ddae9, 18
ALEPH, 0x27702a26126e0b3702af63ee09ac4d1a084ef628, 18
ANT, 0x960b236a07cf122663c4303350609a66a7b288c0, 18
BAL, 0xba100000625a3754423978a60c9317c58a424e3d, 18
BAND, 0xba11d00c5f74255f56a5e366f4f77f5a186d7f55, 18
BAT, 0x0d8775f648430679a709e98d2b0cb6250d2887ef, 18
BNB, 0xb8c77482e45f1f44de1745f52c74426c631bdd52, 18
BNT, 0x1f573d6fb3f13d689ff844b4ce37794d79a7ff1c, 18
BUSD, 0x4fabb145d64652a948d72533023f6e7a623c7c53, 18
BZRX, 0x56d811088235f11c8920698a204a5010a788f4b3, 18
CEL, 0xaaaebe6fe48e54f431b0c390cfaf0b017d09d42d, 4
CHERRY, 0x4ecb692b0fedecd7b486b4c99044392784877e8c, 4
COMP, 0xc00e94cb662c3520282e6f5717214004a7f26888, 18
CRO, 0xa0b73e1ff0b80914ab6fe0444e65848c4c34450b, 8
CRV, 0xd533a949740bb3306d119cc777fa900ba034cd52, 18
DAI, 0x6b175474e89094c44da98b954eedeac495271d0f, 18
DXD, 0xa1d65e8fb6e87b60feccbc582f7f97804b725521, 18
ELF, 0xbf2179859fc6d5bee9bf9158632dc51678a4100e, 18
ENJ, 0xf629cbd94d3791c9250152bd8dfbdf380e2a3b9c, 18
ETH, 0xeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee, 18
EWTB, 0x178c820f862b14f316509ec36b13123da19a6054, 18
FTT, 0x50d1c9771902476076ecfc8b2a83ad6b9355a4c9, 18
GNO, 0x6810e776880c02933d47db1b9fc05908e5386b96, 18
GUSD, 0x056fd409e1d7a124bd7017459dfea2f387b6d5cd, 2
JRT, 0x8a9c67fee641579deba04928c4bc45f66e26343a, 18
KNC, 0xdd974d5c2e2928dea5f71b9825b8b646686bd200, 18
LEND, 0x80fb784b7ed66730e8b1dbd9820afd29931aab03, 18
LINK, 0x514910771af9ca656af840dff83e8264ecf986ca, 18
LRC, 0xbbbbca6a901c926f240b89eacb641d8aec7aeafd, 18
MANA, 0x0f5d2fb29fb7d3cfee444a200298f468908cc942, 18
MATIC, 0x7d1afa7b718fb893db30a3abc0cfc608aacfebb0, 18
MKR, 0x9f8f72aa9304c8b593d555f12ef6589cc3a579a2, 18
MLN, 0xec67005c4e498ec7f55e092bd1d35cbc47c91892, 18
MTA, 0xa3bed4e1c75d00fa6f4e5e6922db7261b5e9acd2, 18
NMR, 0x1776e1f26f98b1a5df9cd347953a26dd3cb46671, 18
OCEAN, 0x967da4048cd07ab37855c090aaf366e4ce1b9f48, 18
OMG, 0xd26114cd6ee289accf82350c8d8487fedb8a0c07, 18
PBTC, 0x5228a22e72ccc52d415ecfd199f99d0665e7733b, 18
RARI, 0xfca59cd816ab1ead66534d82bc21e7515ce441cf, 18
REN, 0x408e41876cccdc0f92210600ef50372656052a38, 18
RENBTC, 0xeb4c2781e4eba804ce9a9803c67d0893436bb27d, 8
RENZEC, 0x1c5db575e2ff833e46a2e9864c22f4b22e0b37c2, 8
RPL, 0xb4efd85c19999d84251304bda99e90b92300bd93, 18
RSR, 0x8762db106b2c2a0bccb3a80d1ed41273552616e8, 18
SNX, 0xc011a73ee8576fb46f5e1c5751ca3b9fe0af2a6f, 18
SRM, 0x476c5e26a75bd202a9683ffd34359c0cc15be0ff, 6
STAKE, 0x0ae055097c6d159879521c384f1d2123d1f195e6, 18
SBTC, 0xfe18be6b3bd88a2d2a7f928d00292e7a9963cfc6, 18
SUSD, 0x57ab1ec28d129707052df4df418d58a2d46d5f51, 18
SUSHI, 0x6b3595068778dd592e39a122f4f5a5cf09c90fe2, 18
SWRV, 0xb8baa0e4287890a5f79863ab62b7f175cecbd433, 18
SXP, 0x8ce9137d39326ad0cd6491fb5cc0cba0e089b6a9, 18
TOMOE, 0x05d3606d5c81eb9b7b18530995ec9b29da05faba, 18
UNI, 0x1f9840a85d5af5bf1d1762f925bdaddc4201f984, 18
USDC, 0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48, 6
USDT, 0xdac17f958d2ee523a2206206994597c13d831ec7, 6
WBTC, 0x2260fac5e5542a773aa44fbcfedf7c193bc2c599, 8
WNXM, 0x0d438f3b5175bebc262bf23753c1e53d03432bde, 18
XDCE, 0x41ab1b6fcbb2fa9dced81acbdec13ea6315f2bf2, 18
YFI, 0x0bc529c00c6401aef6d220be8c6ea1667f6ad93e, 18
UMA, 0x04fa0d235c4abf4bcf4787af4cf447de572ef828, 18
QNT, 0x4a220e6096b25eadb88358cb44068a3248254675, 18
ZRX, 0xe41d2489571d322189246dafa5ebde1f4699f498, 18
CORE, 0x62359ed7505efc61ff1d56fef82158ccaffa23d7, 18
CREAM, 0x2ba592f78db6436527729929aaf6c908497cb200, 18
PICKLE, 0x429881672b9ae42b8eba0e26cd9c73711b891ca5, 18
RCN, 0xf970b8e36e23f7fc3fd752eea86f8be8d83375a6, 18
""")


"""
Carbon helper module -- solidity tests
"""
__VERSION__ = "1.0"
__DATE__ = "11/Feb/2023"

from math import sqrt as _sqrt, ceil as _ceil, log2 as _log2

class SolTestBase():
    """
    testing the solidity curve implementation

    :print_lvl:         the minimum level to be printed (default: LVL_WARN)
    :raise_lvl:         the minimum level where to raise (default: LVL_NEVER)
    :store_lvl:         the minimum level where to store (default: LVL_NEVER)
    :print_context:     whether or not to print the context provided
    :**kwargs:          all go into the `self.context` dict

    *use constants LVL_LOG, LVL_WARN, LVL_ERR, LVL_NEVER 
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    
    def __init__(self, print_lvl=None, raise_lvl=None, store_lvl=None, print_context=False, **kwargs):
        if print_lvl is None: print_lvl = self.PRINT_LVL_DEFAULT
        if raise_lvl is None: raise_lvl = self.RAISE_LVL_DEFAULT
        if store_lvl is None: store_lvl = self.STORE_LVL_DEFAULT
        self.print_lvl      = print_lvl
        self.raise_lvl      = raise_lvl
        self.store_lvl      = store_lvl
        self.print_context  = print_context
        self.context        = kwargs
        
    class OutOfBoundsError(Exception): pass
    class OverflowError(OutOfBoundsError): pass
    class UnderflowError(OutOfBoundsError): pass

    class OutOfBoundsWarning(Exception): pass
    class OverflowWarning(OutOfBoundsWarning): pass
    class UnderflowWarning(OutOfBoundsWarning): pass

    UINT256MAX = 1<<256
    UINT256MAX1 = UINT256MAX-1

    @staticmethod
    def bindig(num):
        """binary digits for num >0 [_ceil(_log2(num+1))]"""
        return _ceil(_log2(num+1))
    
    LVL_LOG     = 0
    LVL_WARN    = 1
    LVL_ERR     = 2
    LVL_NEVER   = 9

    PRINT_LVL_DEFAULT = LVL_WARN
    RAISE_LVL_DEFAULT = LVL_NEVER
    STORE_LVL_DEFAULT = LVL_NEVER
    

    LEVELS = {
        LVL_LOG:    "LOG",
        LVL_WARN:   "WARNING",
        LVL_ERR:    "ERROR", 
    }

    _EXCEPTIONS = {
        True: {
            LVL_WARN: OverflowWarning,
            LVL_ERR: OverflowError,
        },
        False: {
            LVL_WARN: UnderflowWarning,
            LVL_ERR: UnderflowError,
        }
    }
    @classmethod
    def _logmsg(cls, level, isoverflow, label, msg):
        """
        composes and error or warning message
        """
        we = cls.LEVELS[level]
        if level == cls.LVL_LOG: uo = "OK"
        else: uo = "OVERFLOW" if isoverflow else "UNDERFLOW"
        return f"[{we}:{uo}:{label.upper()}] {msg}"

    def store_f(self, level, isoverflow, label, msg, context):
        """stores the event"""
        if level < self.store_lvl: return
        raise NotImplementedError("store_f not implemented yet")

    def raise_f(self, level, isoverflow, label, msg=None, context=None):
        """raises the appropriate exception on errors and warnings"""
        if level < self.raise_lvl: return
        exception = self._EXCEPTIONS[isoverflow][level]
        logmsg = self._logmsg(level, isoverflow, label, msg)
        raise exception(logmsg, self.context, context)
        
    def print_f(self, level, isoverflow, label, msg=None, context=None):
        """prints the appropriate message"""
        if level < self.print_lvl: return
        logmsg = self._logmsg(level, isoverflow, label, msg)
        if self.print_context:
            print(logmsg, self.context, context)
        else:
            print (logmsg)

    def print_raise_f(self, level, isoverflow, label=None, msg=None, context=None):
        """calls print_f first, then raise_f"""
        self.print_f(level, isoverflow, label, msg, context)
        self.raise_f(level, isoverflow, label, msg, context)
        
    UNDERFLOW_ERR  = 100
    UNDERFLOW_WARN = 10000
    OVERFLOW_ERR   = UINT256MAX
    OVERFLOW_WARN  = OVERFLOW_ERR/UNDERFLOW_WARN
    
    def check_uint256(self, number, label=None, context=None, log_f=None):
        """
        checks number for uint256 bounds

        :number:    the number to be checked (cast to int)
        :label:     the label associated with this number
        :context:   other context information associated with this number
        :log_f:     the function to call on error (None: raise)
        :returns:   the number as int in the uint256 range
        """
        number = int(number)
        if log_f is None: log_f = self.print_raise_f
        if label is None: label = ""
        d = self.bindig
        
        if number < 0:
            log_f(level=self.LVL_ERR, isoverflow=True, label=label, msg=f"number < 0", context=context)
            return 0
        elif number < self.UNDERFLOW_ERR:
            log_f(level=self.LVL_ERR, isoverflow=False, label=label, msg=f"{d(number)} bits: < underflow threshold", context=context)
        elif number < self.UNDERFLOW_WARN:
            log_f(level=self.LVL_WARN, isoverflow=False, label=label, msg=f"{d(number)} bits: underflow warning", context=context)
        elif number >= self.OVERFLOW_ERR:
            log_f(level=self.LVL_ERR, isoverflow=True, label=label, msg=f"{d(number)} bits: > max", context=context)
        elif number >= self.OVERFLOW_WARN:
            log_f(level=self.LVL_WARN, isoverflow=True, label=label, msg=f"{d(number)} bits: close to max", context=context)
        else:
            log_f(level=self.LVL_LOG, isoverflow=False, label=label, msg=f"{d(number)} bits: ok", context=context)
        return number

    def __call__(self, number, label=None, context=None):
        """alias for check_uint256 with log_f = default"""
        return self.check_uint256(number, label, context)

        


# ------------------------------------------------------------
# Auto generated test file `test_063_CPC.py`
# ------------------------------------------------------------
# source file   = NBTest_063_CPC.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 063
# test comment  = CPC
# ------------------------------------------------------------



from carbon.helpers.stdimports import *
from carbon import ConstantProductCurve as CPC, CarbonOrderUI
plt.style.use('seaborn-dark')
plt.rcParams['figure.figsize'] = [12,6]
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CPC))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonOrderUI))
print_version(require="2.3.3")




# ------------------------------------------------------------
# Test      063
# File      test_063_CPC.py
# Segment   Assertions and testing
# ------------------------------------------------------------
def test_assertions_and_testing():
# ------------------------------------------------------------
    
    c = CPC.from_px(p=2000,x=10, pair="eth/usdc")
    assert c.pair == "ETH/USDC"
    assert c.tknb == c.pair.split("/")[0]
    assert c.tknx == c.tknb
    assert c.tknq == c.pair.split("/")[1]
    assert c.tkny == c.tknq
    assert f"{c.tknb}/{c.tknq}" == c.pair
    print (c.descr)
    
    c = CPC.from_xy(10,20)
    assert c == CPC.from_kx(c.k, c.x)
    assert c == CPC.from_ky(c.k, c.y)
    assert c == CPC.from_xy(c.x, c.y)
    assert c == CPC.from_pk(c.p, c.k)
    assert c == CPC.from_px(c.p, c.x)
    assert c == CPC.from_py(c.p, c.y)
    
    c = CPC.from_px(p=2, x=100, x_act=10, y_act=20)
    assert c.y_max*c.x_min == c.k
    assert c.x_max*c.y_min == c.k
    assert c.p_min == c.y_min / c.x_max
    assert c.p_max == c.y_max / c.x_min
    assert c.p_max >= c.p_min
    
    c = CPC.from_px(p=2, x=100, x_act=10, y_act=20)
    e = 1e-5
    assert 95*c.yfromx_f(x=95) == c.k
    assert 105*c.yfromx_f(x=105) == c.k
    assert 190*c.xfromy_f(y=190) == c.k
    assert 210*c.xfromy_f(y=210) == c.k
    assert not c.yfromx_f(x=90) is None
    assert c.yfromx_f(x=90-e) is None
    assert not c.xfromy_f(y=180) is None
    assert c.xfromy_f(y=180-e) is None
    assert c.dyfromdx_f(dx=-5)
    assert (c.y+c.dyfromdx_f(dx=-5))*(c.x-5) == c.k
    assert (c.y+c.dyfromdx_f(dx=+5))*(c.x+5) == c.k
    assert (c.x+c.dxfromdy_f(dy=-5))*(c.y-5) == c.k
    assert (c.x+c.dxfromdy_f(dy=+5))*(c.y+5) == c.k
    
    c = CPC.from_pkpp(p=100, k=100)
    assert c.p_min == 100
    assert c.p_max == 100
    assert c.p == 100
    assert c.k == 100
    
    c = CPC.from_pkpp(p=100, k=100, p_min=80, p_max=120)
    assert c.p_min == 80
    assert iseq(c.p_max, 120)
    assert c.p == 100
    assert c.k == 100
    

# ------------------------------------------------------------
# Test      063
# File      test_063_CPC.py
# Segment   iseq
# ------------------------------------------------------------
def test_iseq():
# ------------------------------------------------------------
    
    assert iseq("a", "a", "ab") == False
    assert iseq("a", "a", "a")
    assert iseq(1.0, 1, 1.0)
    assert iseq(0,0)
    assert iseq(0,1e-10)
    assert iseq(0,1e-5) == False
    assert iseq(1, 1.00001) == False
    assert iseq(1, 1.000001)
    assert iseq(1, 1.000001, eps=1e-7) == False
    assert iseq("1", 1) == False
    

# ------------------------------------------------------------
# Test      063
# File      test_063_CPC.py
# Segment   CarbonOrderUI integration
# ------------------------------------------------------------
def test_carbonorderui_integration():
# ------------------------------------------------------------
    
    o = CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2500, 3000, 10, 10)
    c = o.as_cpc
    assert o.pair.slashpair == "ETH/USDC"
    assert o.tkn == "ETH"
    assert o.p_start == 2500
    assert o.p_end == 3000
    assert o.p_marg == 2500
    assert o.y == 10
    assert o.yint == 10
    assert c.pair == o.pair.slashpair
    assert c.tknb == o.pair.tknb
    assert c.tknq == o.pair.tknq
    assert c.x_act == o.y
    assert c.y_act == 0
    assert iseq(o.p_start, c.p, c.p_min)
    assert iseq(o.p_end, c.p_max)
    
    o = CarbonOrderUI.from_prices("ETH/USDC", "USDC", 1500, 1000, 1000, 1000)
    c = o.as_cpc
    assert o.pair.slashpair == "ETH/USDC"
    assert o.tkn == "USDC"
    assert o.p_start == 1500
    assert o.p_end == 1000
    assert o.p_marg == 1500
    assert o.y == 1000
    assert o.yint == 1000
    assert c.pair == o.pair.slashpair
    assert c.tknb == o.pair.tknb
    assert c.tknq == o.pair.tknq
    assert c.x_act == 0
    assert c.y_act == o.y
    assert iseq(o.p_start, c.p, c.p_max)
    assert iseq(o.p_end, c.p_min)
    
    o = CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2500, 3000, 10, 7)
    c = o.as_cpc
    assert o.y == 7
    assert iseq(c.x_act, o.y)
    assert iseq(c.y_act, 0)
    assert iseq(o.p_marg, c.p, c.p_min)
    assert iseq(o.p_end, c.p_max)
    
    o = CarbonOrderUI.from_prices("ETH/USDC", "USDC", 1500, 1000, 1000, 700)
    c = o.as_cpc
    assert o.y == 700
    assert iseq(c.x_act, 0)
    assert iseq(c.y_act, o.y)
    assert iseq(o.p_marg, c.p, c.p_max)
    assert iseq(o.p_end, c.p_min)
    

# ------------------------------------------------------------
# Test      063
# File      test_063_CPC.py
# Segment   Charts [NOTEST]
# ------------------------------------------------------------
def notest_charts():
# ------------------------------------------------------------
    
    # ### Chars (x,y)
    
    xr = np.linspace(1,300,200)
    
    # +
    defaults = dict(p=2)
    curves = [
        CPC.from_px(x=100, **defaults),
        CPC.from_px(x=50, **defaults),
        CPC.from_px(x=150, **defaults),
    ]
    for c in curves:
        plt.plot(xr, [c.yfromx_f(x) for x in xr])
    
    plt.ylim((0,1000))
    plt.xlim((0,300))
    plt.grid()
    
    # +
    defaults = dict(p=2, x_act=10)
    curves = [
        CPC.from_px(x=100, **defaults),
        CPC.from_px(x=50, **defaults),
        CPC.from_px(x=150, **defaults),
    ]
    for c in curves:
        plt.plot(xr, [c.yfromx_f(x) for x in xr])
    
    plt.ylim((0,1000))
    plt.xlim((0,300))
    plt.grid()
    
    # +
    defaults = dict(p=2, y_act=20)
    curves = [
        CPC.from_px(x=100, **defaults),
        CPC.from_px(x=50, **defaults),
        CPC.from_px(x=150, **defaults),
    ]
    for c in curves:
        plt.plot(xr, [c.yfromx_f(x) for x in xr])
    
    plt.ylim((0,1000))
    plt.xlim((0,300))
    plt.grid()
    
    # +
    defaults = dict(p=2, x_act=10, y_act=20)
    curves = [
        CPC.from_px(x=100, **defaults),
        CPC.from_px(x=50, **defaults),
        CPC.from_px(x=150, **defaults),
    ]
    for c in curves:
        plt.plot(xr, [c.yfromx_f(x) for x in xr])
    
    plt.ylim((0,1000))
    plt.xlim((0,300))
    plt.grid()
    # -
    # ### Charts (dx, dy)
    
    
    e=1e-5
    dxr = np.linspace(-50+e,50-e,100)
    
    # +
    defaults = dict(p=2)
    curves = [
        CPC.from_px(x=100, **defaults),
        CPC.from_px(x=50, **defaults),
        CPC.from_px(x=150, **defaults),
    ]
    for c in curves:
        plt.plot(dxr, [c.dyfromdx_f(dx) for dx in dxr])
    
    plt.ylim((-100,200))
    plt.xlim((-50,50))
    plt.grid()
    
    # +
    defaults = dict(p=2, x_act=10)
    curves = [
        CPC.from_px(x=100, **defaults),
        CPC.from_px(x=50, **defaults),
        CPC.from_px(x=150, **defaults),
    ]
    for c in curves:
        plt.plot(dxr, [c.dyfromdx_f(dx) for dx in dxr])
    
    plt.ylim((-100,200))
    plt.xlim((-50,50))
    plt.grid()
    
    # +
    defaults = dict(p=2, y_act=20)
    curves = [
        CPC.from_px(x=100, **defaults),
        CPC.from_px(x=50, **defaults),
        CPC.from_px(x=150, **defaults),
    ]
    for c in curves:
        plt.plot(dxr, [c.dyfromdx_f(dx) for dx in dxr])
    
    plt.ylim((-100,200))
    plt.xlim((-50,50))
    plt.grid()
    
    # +
    defaults = dict(p=2, x_act=10, y_act=20)
    curves = [
        CPC.from_px(x=100, **defaults),
        CPC.from_px(x=50, **defaults),
        CPC.from_px(x=150, **defaults),
    ]
    for c in curves:
        plt.plot(dxr, [c.dyfromdx_f(dx) for dx in dxr])
    
    plt.ylim((-100,200))
    plt.xlim((-50,50))
    plt.grid()
    
    # +
    defaults = dict(p=2, x_act=10, y_act=20)
    curves = [
        CPC.from_px(x=100, **defaults),
        CPC.from_px(x=50, **defaults),
        CPC.from_px(x=150, **defaults),
    ]
    for c in curves:
        plt.plot(dxr, [c.dyfromdx_f(dx) for dx in dxr])
    
    # plt.ylim((-100,200))
    # plt.xlim((-50,50))
    plt.grid()
    # -
    
    


"""
Carbon helper module - efficiently store Carbon coefficients in ints
"""
__VERSION__ = "1.0"
__DATE__ = "10/Feb/2023"

from dataclasses import dataclass as _dataclass, asdict as _asdict
from math import log2 as _log2, ceil as _ceil

@_dataclass
class CarbonFloatInt64():
    """
    represents an integer as significant * 2**exponent with scale one_exponent
    
    :significant:     the significant digits of the number
    :exponent:        the (positive) base-2 exponent of the number
    
    purpose of this class is to compress an integer that has a natural precision
    compoment and a natural scale component; with the default values of 40, 8
    the resulting in can be any integer number with up to 40 significant bits 
    that can be shifted right up to 256 bit, for a total number of bits of
    296 bits.
    
    The global scale (for representing floats) is provided in the class level
    constant ONE_EXPONENT which, as indicated in the class name, is 64; 
    in order to change it use class inheritance
    
        class CarbonFloatInt32(CarbonFloatInt64):
            ONE_EXPONENT = 32
            
    """
    significant: int
    exponent: int

    __VERSION__ = __VERSION__
    __DATE__ = __DATE__
    
    BITS_SIGNIFICANT =  40
    BITS_EXPONENT    =   8
    ONE_EXPONENT     =  64  # the exponent of "number 1", ie the global scale parameter
        
    def __post_init__(self):
        assert isinstance(self.significant, int), f"significant must be an int [{self.significant}]"
        assert isinstance(self.exponent, int), f"exponent must be an int [{self.exponent}]"
        assert self.exponent >= 0, f"exponent must be >= 0 [{self.exponent}]"
    
    @staticmethod
    def bindig(num):
        """binary digits for num >0 [_ceil(_log2(rawint+1))]"""
        return _ceil(_log2(num+1))

    @classmethod
    def from_int(cls, rawint):
        """
        alternative constructor: from raw int
        
        :rawint:    the _actual_ intval to be represented in the class
        """
        assert isinstance(rawint, int), f"value must be int [{rawint}]"
        assert rawint >= 0, f"value must be positive [{rawint}]"
        if rawint == 0:
            return cls(0,0)
        
        bindig = cls.bindig(rawint)
        exponent = bindig - cls.BITS_SIGNIFICANT
        assert exponent <= 2**cls.BITS_EXPONENT, f"exponent must be <= {2**cls.BITS_EXPONENT}, is {exponent} [{intval}]"
        if exponent <= 0:
            return cls(rawint, 0)
        return cls(rawint >> exponent, exponent)
    
    @classmethod
    def from_float(cls, floatval):
        """
        alternative constructor: from float
        
        :floatval:       the number to be represented
        :returns:        cls.from_int(floatval * cls._one())
        """
        return cls.from_int( int(floatval * cls._one()) )
    
    @classmethod
    def max(cls, minus=0):
        """returns maximum possible number (possibly 1,2,3 etc before)"""
        return cls(2**cls.BITS_SIGNIFICANT-1-minus, 2**cls.BITS_EXPONENT-1)
    
    @classmethod
    def min(cls, plus=0):
        """returns minimum possible number > 0 (possibly 1,2,3 etc after)"""
        return cls(1+plus, 0)   
    
    @property
    def astuple(self):
        """alias for tuple(asdict(self).values())"""
        return tuple(self.asdict.values())
    
    @property
    def astuple1(self):
        """alias for tuple(asdict1(self).values())"""
        return tuple(self.asdict1.values())
    
    @property
    def asdict(self):
        """alias for asdict(self)"""
        return _asdict(self)

    @property
    def asdict1(self):
        """adds ONE_EXPONENT to the dict"""
        return {**self.asdict, "one_exponent": self.ONE_EXPONENT}
    
    @property
    def ashex(self):
        """returns the rawint as hex"""
        return hex(self.asint)
    
    @property
    def asint(self):
        """returns the rawint"""
        return self.significant << self.exponent
    
    @property
    def asint1(self):
        """returns the tuple(rawint, 2**one_exponent)"""
        return (self.significant << self.exponent, self.one)
    
    @classmethod
    def _one(cls):
        """returns the global (class level) scaling factor 2**cls.ONE_EXPONENT """
        return 1 << cls.ONE_EXPONENT
    
    @property
    def one(self):
        """wraps classmethod cls._one() as property"""
        return self._one()
    
    @property
    def asfloat(self):
        """returns the float number TAKING INTO ACCOUNT the global exponent `self.one`"""
        return float(self.asint / self.one)
    
    def __int__(self):
        return self.asint
    
    def __float__(self):
        return self.asfloat

class CarbonFloatInt32(CarbonFloatInt64):
    ONE_EXPONENT = 32
    
class CarbonFloatInt40(CarbonFloatInt64):
    ONE_EXPONENT = 40

class CarbonFloatInt48(CarbonFloatInt64):
    ONE_EXPONENT = 48

class CarbonFloatInt60(CarbonFloatInt64):
    ONE_EXPONENT = 60

class CarbonFloatInt64(CarbonFloatInt64):
    ONE_EXPONENT = 64

class CarbonFloatInt128(CarbonFloatInt64):
    ONE_EXPONENT = 128


"""
Carbon helper module - standard Carbon imports for sims

USAGE

    from carbon.helper.stdimports import *

This will print version numbers of key Carbon modules
"""
__VERSION__ = "1.1"
__DATE__ = "25/01/2023"

from math import sqrt, exp, log
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import os

print("[stdimports] imported np, pd, plt, os, sqrt, exp, log")

from .. import CarbonSimulatorUI, CarbonOrderUI, analytics as cal, P, __version__, __date__
from .version import require_version
from . import print_version, plt_style

def iseq(arg0, *args, eps=1e-6):
    """checks whether all arguments are equal to arg0, within tolerance eps if numeric"""
    if not args:
        raise ValueError("Must provide at least one arg", args)
    try:
        arg0+1
        isnumeric = True
    except:
        isnumeric = False
    #if isinstance(arg0, int) or isinstance(arg0, float):
    if isnumeric:
        # numeric testing
        if arg0 == 0:
            for arg in args:
                if abs(arg) > eps: 
                    return False
                return True
        for arg in args:
            if abs(arg/arg0-1) > eps:
                return False
            return True
    else:
        for arg in args:
            if not arg == arg0:
                return False
        return True




# ------------------------------------------------------------
# Auto generated test file `test_057_FloatInt.py`
# ------------------------------------------------------------
# source file   = NBTest_057_FloatInt.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 057
# test comment  = FloatInt
# ------------------------------------------------------------



from carbon.helpers.stdimports import *
from carbon.helpers.floatint import *
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonFloatInt32))
from math import sqrt, log2, ceil, floor
print_version(require="2.3.2")




# ------------------------------------------------------------
# Test      057
# File      test_057_FloatInt.py
# Segment   floatint params
# ------------------------------------------------------------
def test_floatint_params():
# ------------------------------------------------------------
    
    CFI_t = (CarbonFloatInt32, CarbonFloatInt40, CarbonFloatInt48, 
            CarbonFloatInt60, CarbonFloatInt64, CarbonFloatInt128)
    
    assert CarbonFloatInt32.ONE_EXPONENT  == 32
    assert CarbonFloatInt40.ONE_EXPONENT  == 40
    assert CarbonFloatInt48.ONE_EXPONENT  == 48
    assert CarbonFloatInt60.ONE_EXPONENT  == 60
    assert CarbonFloatInt64.ONE_EXPONENT  == 64
    assert CarbonFloatInt128.ONE_EXPONENT == 128
    
    for CFI in CFI_t:
        assert CFI.BITS_SIGNIFICANT == 40
        assert CFI.from_int(2**40-1).exponent == 0
        assert CFI.from_int(2**40).exponent == 1
    

# ------------------------------------------------------------
# Test      057
# File      test_057_FloatInt.py
# Segment   floatint
# ------------------------------------------------------------
def test_floatint():
# ------------------------------------------------------------
    
    I32 = CarbonFloatInt32
    I40 = CarbonFloatInt40
    I64 = CarbonFloatInt64
    
    r = I64.from_float(sqrt(2))
    assert r.significant == 777472127993
    assert r.exponent == 25
    assert r.one == 18446744073709551616
    assert r.asint == 26087635650636414976
    assert r.asint1 == (r.asint, r.one)
    assert r.astuple == (777472127993, 25)
    assert r.astuple1 == (777472127993, 25, 64)
    assert r.astuple1 == (r.significant, r.exponent, r.ONE_EXPONENT)
    assert r.asdict == {'significant': 777472127993, 'exponent': 25}
    assert r.asdict1 == {'significant': 777472127993, 'exponent': 25, 'one_exponent': 64}
    assert r.asdict1 == {'significant': r.significant, 'exponent': r.exponent, 'one_exponent': r.ONE_EXPONENT}
    assert r.asfloat == float(r)
    assert r.asint == int(r)
    
    r  = I64.from_int(2**234-1)
    r2 = I32.from_int(2**234-1)
    assert r.asint == 27606985387137146742797476726052758652111605223017136798872240757145600
    assert r.asint == r2.asint
    assert r.ashex == "0x3fffffffffc000000000000000000000000000000000000000000000000"
    assert r.ashex == r2.ashex
    
    assert r.max().asint == 63657374260394794151270269665081282040354043348133872840046568760304686825600315503411200
    assert r.max().ashex == '0x7fffffffff8000000000000000000000000000000000000000000000000000000000000000'
    assert abs(r.max().asfloat/3.45087e+69-1)<1e-6
    assert r.max().significant == 2**40-1
    assert r.max().exponent == 2**8-1
    
    assert r.min().asint == 1
    assert r.min().ashex == '0x1'
    assert abs(r.min().asfloat/5.421010e-20-1) < 1e-6
    assert r.min().significant == 1
    assert r.min().exponent == 0
    
    assert abs(I32.from_float(sqrt(2)).asfloat/sqrt(2)-1) < 1e-9
    assert abs(I40.from_float(sqrt(2)).asfloat/sqrt(2)-1) < 1e-11
    assert abs(I64.from_float(sqrt(2)).asfloat/sqrt(2)-1) < 1e-11
    
    
    class I96(I64):
        ONE_EXPONENT = 96
    r = I96.from_float(sqrt(2))
    assert r.asint == 112045541949447083908604624896
    assert r.astuple1 == (777472127993, 57, 96)
    
    
    
    


import requests
from math import log10, floor
from dataclasses import dataclass
from . import __VERSION__, __DATE__

class SDKBase:
    """
    base class for SDK access via http

    :url:           RPC URL
    :token:         API token
    :apibase:       API base path
    :syncdefault:   if True, by default call the sync endpoint
    :raiseonerror:  if True, serious errors raise an exception
    :disclaimer:    if True, print a disclaimer on instantiation
    :verbose:       if True, print messages when connecting
    :Tokens:        Token container object for token lookup
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__
    
    URL = None              # to be set in derived class
    TOKEN = None            # to be set in derived class
    APIBASE = None          # to be set in derived class
    SYNCDEFAULT = True      # if True, by default call the sync endpoint
    RAISEONERROR = True     # if True, serious errors raise an exception
    DISCLAIMER = True       # if True, print a disclaimer on instantiation
    VERBOSE = True          # if True, print messages when connecting
    
    def __init__(self, url=None, token=None, apibase=None, syncdefault=None, 
                        raiseonerror=None, disclaimer=None, verbose=None, Tokens=None):
        if url is None: url = self.URL
        self.url = url
        if token is None: token = self.TOKEN
        self.token = token
        if apibase is None: apibase = self.APIBASE
        self.apibase = apibase
        if syncdefault is None: syncdefault = self.SYNCDEFAULT
        self.syncdefault = syncdefault
        if raiseonerror is None: raiseonerror = self.RAISEONERROR
        self.raiseonerror = raiseonerror
        if disclaimer is None: disclaimer = self.DISCLAIMER
        self.disclaimer = disclaimer
        if verbose is None: verbose = self.VERBOSE
        self.verbose = verbose
        self.Tokens = Tokens
        if self.disclaimer:
            print(self.DISCLAIMERTEXT)
        
    POST = "post"
    GET  = "get"

    DISCLAIMERTEXT = """
    ================================================================
    WARNING: DO NOT USE IN PRODUCTION

    This is a demo and testing wrapper for the Carbon SDK. YOU MUST 
    NOT USE THIS SDK, OR ITS ASSOCIATED NODEJS CODE IN PRODUCTION 
    WHEN FUNDS ARE AT RISK. See the disclaimer on the Carbon SDK 
    NodeJS server for more information.
    ================================================================
    """

    class CarbonSDKValueError(ValueError): pass
    
    def _assertTokensProvided(self):
        """raise an exception if Tokens not in SDK; returns Tokens"""
        if self.Tokens is None:
            raise self.CarbonSDKValueError("Tokens not provided in constructor to SDK", self)
        return self.Tokens
    
    @dataclass
    class ErrorResponse():
        """
        dummy class for error responses from the SDK [returns 900 status code]
        """
        error: str

        @property
        def status_code(self):
            return 900
        
        def json(self):
            return {"status_code": self.status_code, "error": self.error}


    def req0(self, path, params=None, method=None):
        """
        execute a GET or POST request to the API server

        :path:      the full path of the call (everything after the server URL)
        :params:    the params transmitted to the API (POST only)
        :method:    self.GET or self.POST (default GET with no params, POST with)
        """
        if method is None: 
            method = self.GET if params is None else self.POST
            #print ("[req0] method is None, setting to", method, params)
        if params is None: params = dict()
        if method == self.GET and params:
            raise self.CarbonSDKValueError("Must not provide params for GET request", params, method)

        url = self.join(self.URL, path)
        if self.verbose:
            print(f"[req0] method={method}, url={url}, params={params}")
        if method == self.GET:
            try:
                return requests.get(url, headers={"token": self.token})
            except Exception as e:
                return self.ErrorResponse(error=str(e))
            
        elif method == self.POST:
            try:
                return requests.post(url, headers={"token": self.token}, json=params)
            except Exception as e:
                return self.ErrorResponse(error=str(e))
        else:
            raise self.CarbonSDKValueError("Unknown method", method)
    
    def req(self, ep, params=None, method=None):
        """
        execute a GET or POST request to the actual API

        :ep:        API endpoint, including the sync/async prefix (eg 'scall/plus')
        :params:    params transmitted to the API (POST only)
        :method:    self.GET or self.POST
        """
        return self.req0(self.join(self.apibase, ep), params, method)
    
    class APICallError(RuntimeError): pass
    class AsyncAPICallError(ValueError): pass

    def call(self, ep, params=None, method=None, sync=None):
        """
        execute a GET or POST request to the (sync or async) call API 

        :ep:        API endpoint, with the sync/async prefix (eg 'plus')
        :params:    params transmitted to the API (POST only)
        :method:    self.GET or self.POST
        :sync:      whether to call sync or async
        """
        url = self.join(self.callapi0(sync), ep)
        if sync: 
            return self.req(url, params, method)
        try:
            r = self.req(url, params, method)
            rj = r.json()
        except:
            raise self.APICallError("Error calling API", r)

        if not rj["success"]:
            if self.raiseonerror:
                raise self.AsyncAPICallError("posts", ep, params, rj)
            return False
        return rj["reqid"]

    def callapi0(self, sync=None):
        """returns "s" if  sync, else "as" """ 
        if sync is None: sync = self.syncdefault
        return "scall" if sync else "ascall"
    
    @staticmethod
    def join(url, path):
        """joins a URL (or path fragment and a path, ensuring exactly one slash in-between"""
        if url[-1] == "/": url = url[:-1]
        if path[0] == "/": path = path[1:]
        return f"{url}/{path}"
    
    @staticmethod
    def c2s(camelCaseString):
        """converts camelCaseString to snake_case_string"""
        return ''.join(['_' + i.lower() if i.isupper() else i for i in camelCaseString]).lstrip('_')
    
    @staticmethod
    def s2c(snakeCaseString):
        """converts snake_case_string to camelCaseString"""
        f = snakeCaseString.split('_')
        return ''.join([f[0]]+[i.title() for i in f[1:]])
    
    @staticmethod
    def roundsd(x, nsd=6):
        """
        rounds the value to significant decimals (returns x unmodified if not a number or zero)
        
        :x:      the value to be rounded
        :nsd:    number of significant decimals
        """
        try:
            magnitude = floor(log10(x))
        except:
            return x
        if magnitude <= nsd:
            return round(x,nsd-magnitude-1)
        else:
            shift = 10**(magnitude-nsd)
            return shift*round(x/shift,0)
  
    @staticmethod
    def bn2int(item):
        """
        converts BigNumber dict to int (if item is a BND; else returns item)

        :returns: int(item) if bn BigNumber dict, else item
        """
        #print("[bn2int]", item)
        if not isinstance(item, dict): return item
        if not item.get("type") == 'BigNumber': return item
        #print("[bn2int] isBn", item)
        return int(item["hex"], 16) 
        
    @classmethod
    def bn2intd(cls, dct):
        """
        converts dict with bignumber entries converted to int entries; other items remain
        """
        #print("[bn2intd]", dct)
        return {k:cls.bn2int(v) for (k,v) in dct.items()}

    @staticmethod
    def int2str(item):
        """
        converts all instances of int to str (if item is int; else returns item)
        """
        if isinstance(item, int): return str(item)
        return item

    @classmethod
    def int2strd(cls, dct):
        """
        converts all instances of int to str in dict; other items remain
        """
        return {k:cls.int2str(v) for (k,v) in dct.items()}

    @staticmethod
    def int2bn(i):
        """
        converts int (or float) to bignumber dict
        """
        return {"type": "BigNumber", 'hex': hex(int(i))}
    
    @staticmethod
    def str2floatd(dct):
        """
        converts all instances of str to float in dict; other items remain
        """
        return {k:float(v) if isinstance(v,str) else v for (k,v) in dct.items()}



"""
Carbon helper module - variables with shared state
"""
import math as _math

__VERSION__ = "1.1"
__DATE__= "28/Feb/2023"

class SharedVar:
    """
    encapsulates numbers into objects so that state can be shared

    :value:      the value with which to initialise the object
    :oid:        disregarded; just for the benefit of __repr__

    see https://rszalski.github.io/magicmethods/

    NOTES

    Essentially this class defines a mutable number. For example

    x = y = ShareVar(5)
    y += 1
    print(x,y) # 6,6
    x.setval(-1)
    print(x,y) # -1,-1
    y = 2
    print(x,y) # -1,2 (link is broken!)    
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__
    
    def __init__(self, value, oid=None):
        self._value = value
    
    @property
    def id(self):
        """the unique object id"""
        return id(self)
     
    @property
    def value(self):
        """the object value"""
        return self._value

    def setval(self, newval):
        """sets the object value to `newval`"""
        self._value = newval
        return self
        
    # comparison

    def __eq__(self, other):
        return self._value == other

    def __eq__(self, other):
        return self._value == other

    def __nq__(self, other):
        return self._value != other

    def __ge__(self, other):
        return self._value >= other

    def __le__(self, other):
        return self._value <= other

    def __lt__(self, other):
        return self._value < other

    def __gt__(self, other):
        return self._value > other

    # unitary operations (return values, not the object)
    def _wrap(self, val):
        """result wrapper"""
        return self.__class__(val)

    def __pos__(self):
        return self._wrap(self._value)

    def __neg__(self):
        return self._wrap(-self._value)

    def __abs__(self):
        return self._wrap(abs(self._value))

    def __round__(self, n):
        return self._wrap(round(self._value,n))

    def __floor__(self):
        return self._wrap(_math.floor(self._value))

    def __ceil__(self):
        return self._wrap(_math.ceil(self._value))

    def __trunc__(self):
        return self._wrap(_math.trunc(self._value))
    
    
    # binary operations

    def __add__(self, other):
        return self._wrap(self._value+other)
    __radd__ = __add__

    def __sub__(self, other):
        return self._wrap(self._value-other)    
    def __rsub__(self, other):
        return self._wrap(other-self._value)

    def __mul__(self, other):
        return self._wrap(self._value*other)
    __rmul__ = __mul__
    
    def __floordiv__(self, other):
        return self._wrap(self._value//other)
    def __rfloordiv__(self, other):
        return self._wrap(other//self._value)

    def __div__(self, other):
        return self._wrap(self._value/other)
    def __rdiv__(self, other):
        return self._wrap(other/self._value)

    def __truediv__(self, other):
        return self._wrap(self._value/other)
    def __rtruediv__(self, other):
        return self._wrap(other/self._value)

    def __mod__(self, other):
        return self._wrap(self._value%other)
    def __rmod__(self, other):
        return self._wrap(other%self._value)

    def __divmod__(self, other):
        d,m = divmod(self._value,other)
        return (self._wrap(d), self._wrap(m))
    def __rdivmod__(self, other):
        d,m = divmod(other, self._value)
        return (self._wrap(d), self._wrap(m))

    def __pow__(self, other):
        return self._wrap(self._value**other)
    def __rpow__(self, other):
        return self._wrap(other**self._value)

    def __lshift__(self, other):
        return self._wrap(self._value<<other)
    def __rlshift__(self, other):
        return self._wrap(other<<self._value)

    def __rshift__(self, other):
        return self._wrap(self._value>>other)
    def __rrshift__(self, other):
            return self._wrap(other>>self._value)
  
    def __and__(self, other):
        return self._wrap(self._value&other)    
    __rand__ = __and__
    
    def __or__(self, other):
        return self._wrap(self._value|other)    
    __ror__ = __or__
    
    def __xor__(self, other):
        return self._wrap(self._value^other)    
    __rxor__ = __xor__

    
    # Assignment operations
    def __iadd__(self, other):
        self._value+=other
        return self
    
    def __isub__(self, other):
        self._value-=other
        return self
        
    def __imul__(self, other):
        self._value*=other
        return self
        
    def __ifloordiv__(self, other):
        self._value//=other
        return self
        
    def __idiv__(self, other):
        self._value//=other
        return self
        
    def __itruediv__(self, other):
        self._value/=other   
        return self
        
    def __imod__(self, other):
        self._value%=other   
        return self
        
    def __ipow__(self, other):
        self._value**=other   
        return self
        
    def __ilshift__(self, other):
        self._value<<=other   
        return self
        
    def __irshift__(self, other):
        self._value>>=other   
        return self
        
    def __iand__(self, other):
        self._value&=other   
        return self
        
    def __ior__(self, other):
        self._value|=other   
        return self
        
    def __ixor__(self, other):
        self._value^=other   
        return self
 

    # Type conversions
    def __int__(self):
        return self.value.__int__()
    
    def __str__(self):
        return self.value.__str__()

    def __long__(self):
        return self.value.__long__()

    def __float__(self):
        return self.value.__float__()

    def __complex__(self):
        return complex(self.value)
    
    def __oct__(self):
        return self.value.__oct__()

    def __hex__(self):
        return self.value.__hex__()

    def __index__(self):
        return self.value.__int__()

    def __coerce__(self):
        return None

    # repr
    def __repr__(self):
        return f"{self.__class__.__name__}({self._value}, oid={self.id})"


# ------------------------------------------------------------
# Auto generated test file `test_059_CurveTest2.py`
# ------------------------------------------------------------
# source file   = NBTest_059_CurveTest2.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 059
# test comment  = CurveTest2
# ------------------------------------------------------------



from carbon.helpers.stdimports import *
from carbon.helpers.soltest import SolTestBase
from carbon.helpers.floatint import *
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(SolTestBase))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonFloatInt32))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonOrderUI))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(P))
from math import log2, floor, ceil, sqrt
print_version(require="2.3.3")


#
#
#
#
#
#
#
#
#
#
#
#
#
#


mulDivF = lambda x, y, z: int((x * y) // z)
mulDiv  = mulDivF
mulDivC = lambda x, y, z: int((x * y + z - 1) // z)


def dy_from_dx(params, C):
    """
    calculates trade amount dx (received by AMM) from dy (sold by AMM)
    
    :params:   tuple(dx, (y,z,A,B,S))
    :C:        the check function C(num, label); must return int(num)
    :returns:  dy
    """

    dx = params[0]
    y,z,A,B,ONE = params[1]
    temp1 = C(y * A + z * B, "temp1")               
    temp2 = C(temp1 * dx / ONE, "temp2")            
    temp3 = C(temp2 * A + z * z * ONE, "temp3") 
    dy = mulDiv(temp1, temp2, temp3)
    #print(temp1, temp2, temp3, dy)
    return dy


def dx_from_dy(params, C):
    """
    calculates trade amount dy (sold by AMM) from dx (received by AMM) 
    
    :params:   tuple(dy, (y,z,A,B,S))
    :C:        the check function C(num, label); must return int(num)
    :returns:  dx
    """

    dy = params[0]
    y,z,A,B,ONE = params[1]
    temp1 = C(z * ONE, "temp1")                  
    temp2 = C(y * A + z * B, "temp2")         
    temp3 = C(temp2 - dy * A, "temp3")        
    scale = mulDiv(temp2, temp3, 2**255)+1
    temp1s = C(temp1//scale, "temp1s")
    temp2s = C(temp2//scale, "temp2s")
    dx = mulDiv(
        C(dy*temp1s, "dx*temp1s"), 
        temp1, 
        C(temp2s*temp3, "temp2s*temp3")
    )
    #print(temp1, temp2, temp3, temp1s, temp2s, dy)
    return dx



# ------------------------------------------------------------
# Test      059
# File      test_059_CurveTest2.py
# Segment   DummyTest
# ------------------------------------------------------------
def test_dummytest():
# ------------------------------------------------------------
    
    assert True
    
    

# ------------------------------------------------------------
# Test      059
# File      test_059_CurveTest2.py
# Segment   Running comparison simple examples [NOTEST]
# ------------------------------------------------------------
def notest_running_comparison_simple_examples():
# ------------------------------------------------------------
    
    class STB(SolTestBase):
        PRINT_LVL_DEFAULT = SolTestBase.LVL_LOG
        #PRINT_LVL_DEFAULT = SolTestBase.LVL_WARN
    VERBOSE = True
    
    # ### SHIB/BTC (Selling BTC)
    
    PAIR = P("SHIB/BTC").sd(18,8)
    price = 1e-5 * 1e-5 # SHIB per BTC
    capacity = 1000 # 1000 BTC
    oui = CarbonOrderUI.from_prices(PAIR, "BTC", price, price/1.05, capacity, capacity)
    curve = oui.yzABS(sx=48, verbose=True)
    curve
    
    params  = (1*1e8, curve)       # dy = BTC-wei
    dx = dx_from_dy( params, STB() )
    dx/1e18 # 1e10 SHIB <- 1 BTC
    
    dx2 = oui.dxfromdy_f(1)
    print(f"dx={dx/1e18:0.2f} vs {dx2:0.2f} {oui.tknx}, diff={abs(dx/1e18/dx2-1)*100:0.4f}%")
    
    params = (1e10*1e18, curve)      # dx = SHIB-wei
    dy = dy_from_dx(params, STB())
    dy/1e8 # 1e3 SHIB <- 1 BTC
    
    dy2 = oui.dyfromdx_f(1e10)
    print(f"dy={dy/1e8:0.4f} vs {dy2:0.4f} {oui.tkny}, diff={abs(dy/1e8/dy2-1)*100:0.4f}%")
    
    # ### SHIB/BTC (selling SHIB)
    
    PAIR = P("SHIB/BTC").sd(18,8)
    price = 1e-5 * 1e-5 # SHIB per BTC # 1e10 USD
    capacity = 1000 * 1e10 # 1000 BTC
    oui = CarbonOrderUI.from_prices(PAIR, "SHIB", price, price*1.05, capacity, capacity)
    curve = oui.yzABS(sx=40, verbose=True)
    curve
    
    params  = (1*1e5*1e5*1e18, curve)       # dy = SHIB-wei (1USD)
    dx = dx_from_dy( params, STB() )
    dx/1e8 # 1e10 SHIB = 1 BTC
    
    dx2 = oui.dxfromdy_f(1)
    print(f"dx={dx/1e8:0.6f} vs {dx2:0.6f} {oui.tknx}, diff={abs(dx/1e18/dx2-1)*100:0.4f}%")
    
    params = (1e8, curve)      # dx = BTC-wei
    dy = dy_from_dx( params, STB() )
    dy/1e18/1e10 #1e10 SHIB = 1 BTC
    
    dy2 = oui.dyfromdx_f(1)
    print(f"dy={dy/1e18:0.4f} vs {dy2:0.4f} {oui.tkny}, diff={abs(dy/1e18/dy2-1)*100:0.8f}%")
    
    

# ------------------------------------------------------------
# Test      059
# File      test_059_CurveTest2.py
# Segment   Running the whole tests [NOTEST]
# ------------------------------------------------------------
def notest_running_the_whole_tests():
# ------------------------------------------------------------
    
    class STB(SolTestBase):
        #PRINT_LVL_DEFAULT = SolTestBase.LVL_LOG
        #PRINT_LVL_DEFAULT = SolTestBase.LVL_WARN
        PRINT_LVL_DEFAULT = SolTestBase.LVL_ERR
    VERBOSE = False
    
    # ### Template code
    
    # +
    ## SETUP Y = TKNQ, X = TKNB
    
    # parameters
    decb = 20
    decq = 1
    sx=48
    amty, amtx = 1, 1
    capacity = 1e10
    width = 0.1
    
    # set up curves y = TKNQ, x = TKNB
    PAIR = P("TKNB/TKNQ").sd(decb,decq)
    price = 1 # TKNQ per TKNB
    capacity = 1e6 # y = 1e6 TKNQ
    oui = CarbonOrderUI.from_prices(PAIR, "TKNQ", price, price/(1+width), capacity, capacity)
    curve = oui.yzABS(sx=sx, verbose=VERBOSE)
    #print(curve)
    
    # dy from dx
    params  = (amty*10**decq, curve)             # amty = TKNQ
    dx  = dx_from_dy(params, STB()) / 10**decb    # dx   = TKNB 
    dx2 = oui.dxfromdy_f(amty)
    print(f"dx={dx:0.6f} vs {dx2:0.6f} {oui.tknx} ==> diff={abs(dx/dx2-1)*100:0.4f}%")
    
    # dx from dy
    params  = (amtx*10**decb, curve)             # amtx = TKNB
    dy = dy_from_dx(params, STB()) / 10**decq    # dy   = TKNQ 
    dy2 = oui.dyfromdx_f(amtx)
    print(f"dy={dy:0.6f} vs {dy2:0.6f} {oui.tkny} ==> diff={abs(dy/dy2-1)*100:0.4f}%")
    dy_from_dx(params, STB())
    # -
    
    # Note: the issues for `decb=0,1` are caused by the fact that the number of tokens expected is very small (1,10 respectively). The float amounts are $10^{-4}$ close to the expected value, but because they are below the int division rounds down, giving 1 token difference, which corresponds to 100% and 10% respectively.
    
    28146000000*9999467920351890/281474989799959504785244160
    
    890090000000*316223491836353792/28147498980863303799508303872
    
    # ### Loop over decimals
    
    # +
    ## SETUP Y = TKNQ, X = TKNB
    
    # parameters
    decb = 20
    sx=40
    amty, amtx = 1, 1
    capacity = 1e12
    width = 0.1
    
    for decq in range(0,41):
        print(f"\ndecq = {decq}")
    
        # set up curves y = TKNQ, x = TKNB
        PAIR = P("TKNB/TKNQ").sd(decb,decq)
        price = 1 # TKNQ per TKNB
        capacity = 1e6 # y = 1e6 TKNQ
        oui = CarbonOrderUI.from_prices(PAIR, "TKNQ", price, price/(1+width), capacity, capacity)
        curve = oui.yzABS(sx=sx, verbose=VERBOSE)
        #print(curve)
    
        # dy from dx
        params  = (amty*10**decq, curve)             # amty = TKNQ
        dx  = dx_from_dy(params, STB()) / 10**decb    # dx   = TKNB 
        dx2 = oui.dxfromdy_f(amty)
        print(f"dx={dx:0.6f} vs {dx2:0.6f} {oui.tknx} ==> diff={abs(dx/dx2-1)*100:0.4f}%")
    
        # dx from dy
        params  = (amtx*10**decb, curve)             # amtx = TKNB
        dy = dy_from_dx(params, STB()) / 10**decq    # dy   = TKNQ 
        dy2 = oui.dyfromdx_f(amtx)
        print(f"dy={dy:0.6f} vs {dy2:0.6f} {oui.tkny} ==> diff={abs(dy/dy2-1)*100:0.4f}%")
    # -
    
    # ### Loop over prices (18/18)
    
    # +
    ## SETUP Y = TKNQ, X = TKNB
    
    # parameters
    decb = 18
    decq = decb
    sx=40
    amtx = 1 # TKNB
    capacity = 1e6
    width = 0.1
    
    for priceexp in range(-10,10):
        price = 10**priceexp
        amty = amtx*price # TKNQ
        print(f"\nprice = {price} TKNQ per TKNB [10**{priceexp}]")
    
        # set up curves y = TKNQ, x = TKNB
        PAIR = P("TKNB/TKNQ").sd(decb,decq)
        #price = 1 # TKNQ per TKNB
        capacity = 1e6 # y = 1e6 TKNQ
        oui = CarbonOrderUI.from_prices(PAIR, "TKNQ", price, price/(1+width), capacity*price, capacity*price)
        curve = oui.yzABS(sx=sx, verbose=VERBOSE)
        #print(curve)
    
        # dy from dx
        params  = (amty*10**decq, curve)             # amty = TKNQ
        dx  = dx_from_dy(params, STB()) / 10**decb    # dx   = TKNB 
        dx2 = oui.dxfromdy_f(amty)
        print(f"dx={dx:0.6f} vs {dx2:0.6f} {oui.tknx} ==> diff={abs(dx/dx2-1)*100:0.4f}%")
    
        # dx from dy
        params  = (amtx*10**decb, curve)             # amtx = TKNB
        dy = dy_from_dx(params, STB()) / 10**decq    # dy   = TKNQ 
        dy2 = oui.dyfromdx_f(amtx)
        print(f"dy={dy:0.6f} vs {dy2:0.6f} {oui.tkny} ==> diff={abs(dy/dy2-1)*100:0.4f}%")
    # -
    
    # ### Loop over prices (18/6)
    
    # +
    ## SETUP Y = TKNQ, X = TKNB
    
    # parameters
    decb = 18
    decq = 6
    sx=48
    amtx = 1 # TKNB
    capacity = 1e6
    width = 0.1
    
    for priceexp in range(-10,10):
        price = 10**priceexp
        amty = amtx*price # TKNQ
        print(f"\nprice = {price} TKNQ per TKNB [10**{priceexp}]")
    
        # set up curves y = TKNQ, x = TKNB
        PAIR = P("TKNB/TKNQ").sd(decb,decq)
        #price = 1 # TKNQ per TKNB
        capacity = 1e6 # y = 1e6 TKNQ
        oui = CarbonOrderUI.from_prices(PAIR, "TKNQ", price, price/(1+width), capacity*price, capacity*price)
        curve = oui.yzABS(sx=sx, verbose=VERBOSE)
        #print(curve)
    
        # dy from dx
        params  = (amty*10**decq, curve)             # amty = TKNQ
        dx  = dx_from_dy(params, STB()) / 10**decb    # dx   = TKNB 
        dx2 = oui.dxfromdy_f(amty)
        print(f"dx={dx:0.6f} vs {dx2:0.6f} {oui.tknx} ==> diff={abs(dx/dx2-1)*100:0.4f}%")
    
        # dx from dy
        params  = (amtx*10**decb, curve)             # amtx = TKNB
        dy = dy_from_dx(params, STB()) / 10**decq    # dy   = TKNQ 
        dy2 = oui.dyfromdx_f(amtx)
        print(f"dy={dy:0.6f} vs {dy2:0.6f} {oui.tkny} ==> diff={abs(dy/dy2-1)*100:0.4f}%")
    # -
    
    


# ------------------------------------------------------------
# Auto generated test file `test_058_CurveTest.py`
# ------------------------------------------------------------
# source file   = NBTest_058_CurveTest.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 058
# test comment  = CurveTest
# ------------------------------------------------------------



from carbon.helpers.stdimports import *
from carbon.helpers.soltest import SolTestBase
from carbon.helpers.floatint import *
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(SolTestBase))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonFloatInt32))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonOrderUI))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(P))
from math import log2, floor, ceil, sqrt
print_version(require="2.3.3")


#
#
#
#
#
#
#
#
#
#

FI32 = CarbonFloatInt32
FI40 = CarbonFloatInt40


# ------------------------------------------------------------
# Test      058
# File      test_058_CurveTest.py
# Segment   Demo and test of yzABS in CarbonOrderUI and pair decimals
# ------------------------------------------------------------
def test_demo_and_test_of_yzabs_in_carbonorderui_and_pair_decimals():
# ------------------------------------------------------------
    
    ETHUSDC = P("ETH/USDC").sd(18,6)
    assert ETHUSDC.decimals == {'ETH': 18, 'USDC': 6, '_TKNB': 18, '_TKNQ': 6, '_DIFFQB': -12}
    assert ETHUSDC.decdiffqb == -12
    assert P(ETHUSDC).has_decimals
    assert ETHUSDC.price_convention == 'USDC per ETH'
    ETHUSDC
    
    oui = CarbonOrderUI.from_prices(ETHUSDC, "ETH", 1500, 2000, 1, 0.5)
    r = oui.yzABS(sx=10, verbose=True)
    ddf, ddf2 = 10**12, 10**6
    assert r.y == 0.5 * 10**18
    assert r.z == 10**18
    assert r.A == int(oui.S * ddf2 * r.S)
    assert r.B == int(oui.B * ddf2 * r.S)
    assert r.S == 2**10
    assert oui.descr() == 'Sell ETH buy USDC from 1500.0000 to 2000.0000 USDC per ETH'
    
    oui = CarbonOrderUI.from_prices(ETHUSDC, "USDC", 1000, 500, 1100, 790)
    r = oui.yzABS(sx=40, verbose=True)
    ddf, ddf2 = 10**-12, 10**-6
    assert r.y == 790 * 10**6
    assert r.z == 1100 * 10**6
    assert r.A == int(oui.S * ddf2 * r.S)
    assert r.B == int(oui.B * ddf2 * r.S)
    assert r.S == 2**40
    

# ------------------------------------------------------------
# Test      058
# File      test_058_CurveTest.py
# Segment   Sundry tests of other carbonui stuff
# ------------------------------------------------------------
def test_sundry_tests_of_other_carbonui_stuff():
# ------------------------------------------------------------
    
    oui = CarbonOrderUI.from_prices(P("ETH/USDC").sd(18,6), "USDC", 1000, 500, 1100, 790)
    r = oui.set_id(1)
    assert oui.tkn == "USDC"
    assert oui.tkny == oui.tkn
    assert oui.tknx == "ETH"
    assert oui.pair.slashpair == "ETH/USDC"
    assert abs(oui.S/9.262096826685895-1) < 1e-10
    assert abs(oui.B/22.360679774997898-1) < 1e-10
    assert oui.S is oui.A
    assert abs(oui.pa/1000-1) < 1e-10
    assert abs(oui.pa_raw/1000-1) < 1e-10
    assert oui.p_start == oui.pa
    assert oui.py == oui.pa
    assert abs(oui.pb/500-1) < 1e-10
    assert abs(oui.pb_raw/500-1) < 1e-10
    assert oui.p_end == oui.pb
    assert oui.py == oui.py
    assert abs(oui.yint/1100-1) < 1e-10
    assert abs(oui.y/790-1) < 1e-10
    assert oui.y == oui.z
    assert oui.total_liquidity == (790, 'USDC')
    assert oui.price_convention == 'USDC per ETH'
    assert oui.price_convention == oui.pair.price_convention
    assert oui.price_convention == oui.price_convention_raw
    assert oui.yzABS(32).astuple == (790000000, 1100000000, 39780, 96038, 4294967296)
    assert r is oui
    try:
        oui.set_id(1)
        raise
    except ValueError as e:
        print(e)
    
    oui.price_convention_raw
    
    oui2 = CarbonOrderUI.from_prices(P("ETH/USDC").sd(18,6), "ETH", 1500, 2000, 2, 1)
    oui2.set_id(2)
    r  = oui.set_linked(oui2)
    r2 = oui2.set_linked(oui)  # sets linked here and creates backlink
    assert oui2.tkn == "ETH"
    assert oui2.tkny == oui2.tkn
    assert oui2.tknx == "USDC"
    assert oui2.pair.slashpair == "ETH/USDC"
    assert abs(oui2.S/0.0034592091997182155-1) < 1e-10
    assert abs(oui2.B/0.022360679774997897-1) < 1e-10
    assert oui2.S is oui2.A
    assert abs(oui2.pa/1500-1) < 1e-10
    assert abs(oui2.pa_raw/0.0006666666666666666-1) < 1e-10
    assert oui2.p_start == oui2.pa
    assert oui2.py == oui2.pa
    assert abs(oui2.pb/2000-1) < 1e-10
    assert abs(oui2.pb_raw/0.0005-1) < 1e-10
    assert oui2.p_end == oui2.pb
    assert oui2.py == oui2.py
    assert abs(oui2.yint/2-1) < 1e-10
    assert abs(oui2.y/1-1) < 1e-10
    assert oui2.y == oui2.z
    assert oui2.total_liquidity == (1, 'ETH')
    assert oui2.price_convention == 'USDC per ETH'
    assert oui2.price_convention == oui2.pair.price_convention
    assert oui2.price_convention != oui2.price_convention_raw
    assert oui2.price_convention_raw == "ETH per USDC"
    assert oui2.yzABS(32).astuple == (1000000000000000000, 2000000000000000000, 14857190382812, 96038388349944, 4294967296)
    assert oui2.yzABS(32).asdict == {
        'y': 1000000000000000000,
        'z': 2000000000000000000,
        'A': 14857190382812,
        'B': 96038388349944,
        'S': 4294967296
    }
    assert r is oui
    assert r2 is oui2
    try:
        oui2.set_linked(oui)
        raise
    except ValueError as e:
        print(e)
    
    oui2.yzABS(32).asdict
    
    assert oui.lid == 2
    assert oui2.lid == 1
    

# ------------------------------------------------------------
# Test      058
# File      test_058_CurveTest.py
# Segment   Demo [NOTEST]
# ------------------------------------------------------------
def notest_demo():
# ------------------------------------------------------------
    
    # ### Trade functions
    
    mulDivF = lambda x, y, z: (x * y) // z
    mulDiv = mulDivF
    mulDivC = lambda x, y, z: (x * y + z - 1) // z
    
    
    # #### trade_by_source_dy_from_dx
    
    def trade_by_source_dy_from_dx(params, C):
    
        dx = params[0]
        y,z,A,B,s = params[1]
        print(params)
        ONE = s
        temp1 = C(y * A + z * B, "temp1")               # 177 bits at most; cannot overflow
        temp2 = C(temp1 * dx / ONE, "temp2")            # 224 bits at most; can overflow; some precision loss
        temp3 = C(temp2 * A + z * z * ONE, "temp3")     # 256 bits at most; can overflow
        dy = mulDiv(temp1, temp2, temp3)
        print(dx, temp1, temp2, temp3, dy)
        return dy
    
    
    # +
    # temp1 = z * ONE
    # temp2 = y * A + z * B
    # temp3 = temp2 * dy
    # scale = mulDivC(temp3, A, 2**256-1)
    # temp4 = mulDivC(temp1, temp1, scale)
    # temp5 = mulDivC(temp3, A, scale)
    # dx    = mulDivF(temp2, temp3 // scale, temp4 + temp5)
    # -
    
    # #### trade_by_target_dx_from_dy
    #
    # this one is SKL version of the fixed code
    
    def trade_by_target_dx_from_dy(params, C):
    
        dy = params[0]
        y,z,A,B,s = params[1]
        print(params)
        ONE = s
        temp1 = C(z * ONE, "temp1")                  
        temp2 = C(y * A + z * B, "temp2")         
        temp3 = C(temp2 - dy * A, "temp3")        
        scale = mulDiv(temp2, temp3, 2**255)+1
        temp1s = C(temp1//scale, "temp1s")
        temp2s = C(temp2//scale, "temp2s")
        dx = mulDiv(
            C(dy*temp1s, "dx*temp1s"), 
            temp1, 
            C(temp2s*temp3, "temp2s*temp3")
        )
        print(dy, dx*temp1s, temp1, temp2s*temp3, dx)
        return dx
    
    
    # this is BM version of the fixed code
    
    # +

    # mulDivF = lambda x, y, z: x * y // z
    # mulDivC = lambda x, y, z: (x * y + z - 1) // z
    
    # temp1 = z * ONE
    # temp2 = y * A + z * B
    # temp3 = temp2 - dx * A
    # scale = mulDivC(temp2, temp3, 2**256-1)
    # temp4 = mulDivC(temp1, temp1, scale)
    # temp5 = mulDivF(temp2, temp3, scale)
    # dy    = mulDivC(dx, temp4, temp5)
    # -
    
    # that's the previous, failing, version of the code
    
    def trade_by_target_dx_from_dy_old(params, C):
    
        dy = params[0]
        y,z,A,B,s = params[1]
        ONE = s
        temp1 = C(z * ONE, "temp1")                  
        temp2 = C(y * A + z * B, "temp2")         
        temp3 = C(temp2 - dy * A, "temp3")        
        dx = mulDiv(
            C(dy*temp1, "dy*temp1"), 
            temp1, 
            C(temp2*temp3, "temp2*temp3")
        ) 
        print(dy, dy*temp1, temp1, temp2*temp3, dx)
        return dx
    
    
    # ### Analysis
    
    class STB(SolTestBase):
        #PRINT_LVL_DEFAULT = SolTestBase.LVL_LOG
        PRINT_LVL_DEFAULT = SolTestBase.LVL_WARN
    VERBOSE = False
    
    # #### TKN/DAI -- same decimality, same reasonable price
    
    TKNDAI = P("TKN/DAI").sd(18,18)
    price = 5  # DAI per TKN
    oui = CarbonOrderUI.from_prices(TKNDAI, "TKN", price, price, 1000, 1000)
    c = curve = oui.yzABS(sx=40, verbose=VERBOSE)
    
    params_bysrc  = (1e18, curve) # dy = wei sent (by source, 1 TKN)
    dx = trade_by_target_dx_from_dy( params_bysrc, STB(context=("by_source", curve)))
    dx/1e18 # 1 TKN -> 5 DAI ==> 5 DAI per TKN
    
    params_bytarg = (5e18, curve)       # dx = token wei received (target, DAI)
    dy = trade_by_source_dy_from_dx( params_bytarg, STB(context=("by_target", curve)) )
    dy/1e18 # 1 TKN ->  5 DAI --> 5 DAI per TKN
    
    # #### ETH/USDC -- y=ETH, x=USDC
    
    ETHUSDC = P("ETH/USDC").sd(18,6)
    oui = CarbonOrderUI.from_prices(ETHUSDC, "ETH", 2000, 2000, 1, 0.5)
    curve = oui.yzABS(sx=40, verbose=VERBOSE)
    
    params_bysrc  = (1*1e18, curve) # dy = token wei sent (by source, 1 ETH)
    dx = trade_by_target_dx_from_dy( params_bysrc, STB(context=("by_source", curve)))
    dx/1e6 # 1 ETH -> 2000 USDC
    
    params_bytarg = (2000*1e6, curve)       # dx = token wei received (target, USDC)
    dy = trade_by_source_dy_from_dx( params_bytarg, STB(context=("by_target", curve)) )
    dy/1e18 # 1 ETH -> 2000 USDC
    
    # #### ETH/USDC -- y=USDC, x=ETH
    
    oui = CarbonOrderUI.from_prices(ETHUSDC, "USDC", 1000, 1000, 750, 750)
    curve = oui.yzABS(sx=40, verbose=VERBOSE)
    print(curve)
    
    params_bysrc  = (1000*1e6, curve)       # dy = USDC-wei
    dx = trade_by_target_dx_from_dy( params_bysrc, STB(context=("by_source", curve)) )
    dx/1e18 # 1000 USDC -> 1 ETH
    
    params_bytarg = (1e18, curve)  # dx = ETH-wei
    dy = trade_by_source_dy_from_dx( params_bytarg, STB(context=("by_target", curve)) )
    dy/1e6 # 1000 USDC -> 1 ETH
    
    # #### SHIB/USDC -- y=SHIB, x=USDC
    
    SHIBUSDC = P("SHIB/USDC").sd(18,6)
    price = 1e-5
    oui = CarbonOrderUI.from_prices(SHIBUSDC, "SHIB", price, price, 2e5, 2e5)
    curve = oui.yzABS(sx=40, verbose=VERBOSE)
    print(curve)
    
    params_bysrc  = (1*1e5*1e18, curve)      # dy = SHIB-wei (1 USD worth of SHIB)
    dx = trade_by_target_dx_from_dy( params_bysrc, STB(context=("by_source", curve)) )
    dx / 1e6 # 1e5 SHIB -> 1 USDC
    
    params_bytarg = (1*1e6, curve) # dx = USDC-wei
    dy = trade_by_source_dy_from_dx( params_bytarg, STB(context=("by_target", curve)) )
    dy/1e18 # 1e5 SHIB -> 1 USDC
    
    # #### Nick's example DAI/USDC -- Selling DAI
    
    DAIUSDC = P("DAI/USDC").sd(18,6)
    price = 1
    oui = CarbonOrderUI.from_prices(DAIUSDC, "DAI", price, price, 1e5, 1e5)
    curve = oui.yzABS(sx=40, verbose=VERBOSE)
    print(curve)
    
    params_bysrc  = (1*1e18, curve)        # dy = DAI-wei
    dx = trade_by_target_dx_from_dy( params_bysrc, STB(context=("by_source", curve)) )
    dx/1e6 # 1 DAI -> 1 USDC
    
    
    params_bytarg = (1*1e6, curve)  # dx = USDC-wei
    dy = trade_by_source_dy_from_dx( params_bytarg, STB(context=("by_target", curve)) )
    dy / 1e18 # 1 DAI -> 1 USDC
    
    
    # #### Nick's example DAI/USDC -- Selling USDC
    
    SHIBUSDC = P("DAI/USDC").sd(18,6)
    price = 1
    oui = CarbonOrderUI.from_prices(SHIBUSDC, "USDC", price, price, 1e5, 1e5)
    curve = oui.yzABS(sx=40, verbose=VERBOSE)
    
    params_bysrc  = (1*1e6, curve)       # dy = USDC-wei
    dx = trade_by_target_dx_from_dy( params_bysrc, STB(context=("by_source", curve)) )
    dx/1e18
    
    params_bytarg = (1e18, curve)      # dx = DAI-wei
    dy = trade_by_source_dy_from_dx( params_bytarg, STB(context=("by_target", curve)) )
    dy/1e6
    
    

# ------------------------------------------------------------
# Test      058
# File      test_058_CurveTest.py
# Segment   More examples [NOTEST]
# ------------------------------------------------------------
def notest_more_examples():
# ------------------------------------------------------------
    
    class STB(SolTestBase):
        PRINT_LVL_DEFAULT = SolTestBase.LVL_LOG
        #PRINT_LVL_DEFAULT = SolTestBase.LVL_WARN
    VERBOSE = True
    
    # ### DAI/USDC
    
    PAIR = P("DAI/USDC").sd(18,6)
    price = 1
    oui = CarbonOrderUI.from_prices(PAIR, "USDC", price, price, 1e5, 1e5)
    curve = oui.yzABS(sx=40, verbose=VERBOSE)
    curve
    
    params  = (1*1e6, curve)       # dy = USDC-wei
    dx = trade_by_target_dx_from_dy( params, STB(context=("by_source", curve)) )
    dx/1e18
    
    params = (1e18, curve)      # dx = DAI-wei
    dy = trade_by_source_dy_from_dx( params, STB(context=("by_target", curve)) )
    dy/1e6
    
    # ### SHIB/USDC
    
    PAIR = P("SHIB/USDC").sd(18,6)
    price = 1e-5 # SHIB per USDC
    oui = CarbonOrderUI.from_prices(PAIR, "SHIB", price, price, 1e10, 1e10)
    curve = oui.yzABS(sx=40, verbose=VERBOSE)
    curve
    
    params  = (1*1e5*1e18, curve)       # dy = SHIB-wei
    dx = trade_by_target_dx_from_dy( params, STB(context=("by_source", curve)) )
    dx/1e6
    
    params = (1e6, curve)      # dx = USDC-wei
    dy = trade_by_source_dy_from_dx( params, STB(context=("by_target", curve)) )
    dy/1e18
    
    # ### SHIB/BTC (selling SHIB)
    
    PAIR = P("SHIB/BTC").sd(18,8)
    price = 1e-5 * 1e-5 # SHIB per BTC # 1e10 USD
    capacity = 1000 * 1e10 # 1000 BTC
    oui = CarbonOrderUI.from_prices(PAIR, "SHIB", price, price*1.05, capacity, capacity)
    curve = oui.yzABS(sx=40, verbose=True)
    curve
    
    # #### Trading 1e10 SHIB -> 1 BTC
    
    params  = (1*1e5*1e5*1e18, curve)       # dy = SHIB-wei (1USD)
    dx = trade_by_target_dx_from_dy( params, STB(context=("by_source", curve)) )
    dx/1e8 # 1e10 SHIB = 1 BTC
    
    params = (1e8, curve)      # dx = BTC-wei
    dy = trade_by_source_dy_from_dx( params, STB(context=("by_target", curve)) )
    dy/1e18/1e10 #1e10 SHIB = 1 BTC
    
    # #### Trading 1e5 SHIB -> 1e-5 BTC (1USD)
    
    params  = (1*1e5*1e18, curve)       # dy = SHIB-wei (1USD)
    dx = trade_by_target_dx_from_dy( params, STB(context=("by_source", curve)) )
    dx/1e8/1e-5 # 1e5 SHIB = 1e-5 BTC
    
    params = (1e-5*1e8, curve)      # dx = BTC-wei
    dy = trade_by_source_dy_from_dx( params, STB(context=("by_target", curve)) )
    dy/1e18/1e5 # 1e5 SHIB = 1e-5 BTC
    
    # #### Trading 1e3 SHIB -> 1e-7 BTC (0.01USD)
    
    params  = (1*1e3*1e18, curve)       # dy = SHIB-wei (1USD cent)
    dx = trade_by_target_dx_from_dy( params, STB(context=("by_source", curve)) )
    dx/1e8/1e-7 # 1e3 SHIB = 1e-7 BTC
    
    params = (1e-7*1e8, curve)      # dx = BTC-wei
    dy = trade_by_source_dy_from_dx( params, STB(context=("by_target", curve)) )
    dy/1e18/1e3 # 1e3 SHIB = 1e-7 BTC
    
    # ### SHIB/BTC (selling BTC)
    
    PAIR = P("SHIB/BTC").sd(18,8)
    price = 1e-5 * 1e-5 # SHIB per BTC
    capacity = 1000 # 1000 BTC
    oui = CarbonOrderUI.from_prices(PAIR, "BTC", price, price/1.05, capacity, capacity)
    curve = oui.yzABS(sx=48, verbose=True)
    curve
    
    # #### Trading 1 BTC -> 1e10 SHIB
    
    params  = (1*1e8, curve)       # dy = BTC-wei
    dx = trade_by_target_dx_from_dy( params, STB(context=("by_source", curve)) )
    dx/1e18/1e10 # 1e10 SHIB <- 1 BTC
    
    params = (1e10*1e18, curve)      # dx = SHIB-wei
    dy = trade_by_source_dy_from_dx( params, STB(context=("by_target", curve)) )
    dy/1e8 # 1e3 SHIB <- 1 BTC
    
    # #### Trading 1e-5 BTC -> 1e5 SHIB (1USD)
    
    params  = (1e-5*1e8, curve)       # dy = BTC-wei
    dx = trade_by_target_dx_from_dy( params, STB(context=("by_source", curve)) )
    dx/1e18/1e5 # 1e5 SHIB <- 1e-5 BTC
    
    params = (1e5*1e18, curve)      # dx = SHIB-wei
    dy = trade_by_source_dy_from_dx( params, STB(context=("by_target", curve)) )
    dy/1e8/1e-5 # 1e3 SHIB <- 1e-5 BTC
    
    # #### Trading 1e-7 BTC -> 1e3 SHIB (0.01USD)
    
    params  = (1e-7*1e8, curve)       # dy = BTC-wei
    dx = trade_by_target_dx_from_dy( params, STB(context=("by_source", curve)) )
    dx/1e18/1e3 # 1e3 SHIB <- 1e-7 BTC
    
    params = (1e3*1e18, curve)      # dx = SHIB-wei
    dy = trade_by_source_dy_from_dx( params, STB(context=("by_target", curve)) )
    dy/1e8/1e-7 # 1e3 SHIB <- 1e-7 BTC
    
    # #### Trading 10 BTC -> 1e11 SHIB
    
    params  = (10*1e8, curve)       # dy = BTC-wei
    dx = trade_by_target_dx_from_dy( params, STB(context=("by_source", curve)) )
    dx/1e18/1e11 # 1e11 SHIB <- 10 BTC
    
    params = (1e11*1e18, curve)      # dx = SHIB-wei
    dy = trade_by_source_dy_from_dx( params, STB(context=("by_target", curve)) )
    dy/1e8/10 # 1e3 SHIB <- 100 BTC
    
    # #### Trading 100 BTC -> 1e12 SHIB
    
    params  = (100*1e8, curve)       # dy = BTC-wei
    dx = trade_by_target_dx_from_dy( params, STB(context=("by_source", curve)) )
    dx/1e18/1e12 # 1e12 SHIB <- 100 BTC
    
    params = (1e12*1e18, curve)      # dx = SHIB-wei
    dy = trade_by_source_dy_from_dx( params, STB(context=("by_target", curve)) )
    dy/1e8/100 # 1e3 SHIB <- 100 BTC
    

# ------------------------------------------------------------
# Test      058
# File      test_058_CurveTest.py
# Segment   SolTestBase tests
# ------------------------------------------------------------
def test_soltestbase_tests():
# ------------------------------------------------------------
    
    TB0 = SolTestBase()
    TBrw = SolTestBase(raise_lvl=TB0.LVL_WARN)
    TBre = SolTestBase(raise_lvl=TB0.LVL_ERR)
    
    try:
        TBrw.check_uint256(1, "testlabel")
        raise RuntimeError("should raise")
    except TB0.UnderflowError as e:
        print(e)
    try:
        TBrw.check_uint256(2000, "testlabel")
        raise RuntimeError("should raise")
    except TB0.UnderflowWarning as e:
        print(e)
    try:
        TBrw.check_uint256(2**254, "testlabel")
        raise RuntimeError("should raise")
    except TB0.OverflowWarning as e:
        print(e)
    try:
        TBrw.check_uint256(2**256-1, "testlabel")
        raise RuntimeError("should raise")
    except TB0.OverflowWarning as e:
        print(e)
    try:
        TBrw.check_uint256(2**256, "testlabel")
        raise RuntimeError("should raise")
    except TB0.OverflowError as e:
        print(e)
    
    
    try:
        TBre.check_uint256(1, label="testlabel")
        raise RuntimeError("should raise")
    except TBre.UnderflowError as e:
        print(e)
    assert TBre.check_uint256(2000, label="testlabel") == 2000
    assert TBre.check_uint256(2**254, label="testlabel") == 2**254
    assert TBre.check_uint256(2**256-1, label="testlabel") == 2**256-1
    try:
        TBre.check_uint256(2**256, label="testlabel")
        raise RuntimeError("should raise")
    except TBre.OverflowError as e:
        print(e)
    
    
    assert TB0.check_uint256(1, log_f=TB0.print_f) == 1
    assert TB0.check_uint256(2000, log_f=TB0.print_f) == 2000
    assert TB0.check_uint256(2**254, log_f=TB0.print_f) == 2**254
    assert TB0.check_uint256(2**256-1, log_f=TB0.print_f) == 2**256-1
    assert TB0.check_uint256(2**256, log_f=TB0.print_f) == 2**256
    
    assert TB0._logmsg(level=TB0.LVL_WARN, isoverflow=True, label="1", msg="") == '[WARNING:OVERFLOW:1] '
    assert TB0._logmsg(TB0.LVL_ERR, False, "mylabel", "mymessage") == '[ERROR:UNDERFLOW:MYLABEL] mymessage'
    TB0.print_f(False, False, "mylabel", "mymessage")
    
    assert TB0.bindig(0) == 0
    assert TB0.bindig(1) == 1
    assert TB0.bindig(2) == 2
    assert TB0.bindig(3) == 2
    assert TB0.bindig(2**10) == 11
    assert TB0.bindig(2**10+1) == 11
    assert TB0.bindig(2**10-1) == 10
    
    